/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** Info: 
**    Example of how to use libnotify correctly and at the same time comply to
**    the new notification spec (read: visual guidelines)
**
** Compile:
**    gcc -O0 -ggdb -Wall -Werror `pkg-config --cflags --libs libnotify \
**    glib-2.0` sd-example-incoming-call.c example-util.c -o \
**    sd-example-incoming-call
**
** Copyright 2012 Canonical Ltd.
**
** Author:
**    Mirco "MacSlow" Mueller <mirco.mueller@canonical.com>
**
** This program is free software: you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 3, as published
** by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranties of
** MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
** PURPOSE.  See the GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program.  If not, see <http://www.gnu.org/licenses/>.
**
*******************************************************************************/

#include "example-util.h"

void
action_decline_1 (NotifyNotification* notification,
                  const char*               action,
                  gpointer            data)
{
	if (!g_strcmp0 (action, "action_decline_1"))
		g_print ("Hang up\n");
	else
		g_print ("That should not have happened (action_decline_1)!");
}

void
action_decline_2 (NotifyNotification* notification,
                  const char*               action,
                  gpointer            data)
{
	if (!g_strcmp0 (action, "action_decline_2"))
		g_print ("\"Dude, my wife is here!\"\n");
	else
		g_print ("That should not have happened (action_decline_2)!");
}

void
action_decline_3 (NotifyNotification* notification,
                  const char*               action,
                  gpointer            data)
{
	if (!g_strcmp0 (action, "action_decline_3"))
		g_print ("\"I'm sleeping.\"\n");
	else
		g_print ("That should not have happened (action_decline_3)!");
}

void
action_decline_4 (NotifyNotification* notification,
                  const char*               action,
                  gpointer            data)
{
	if (!g_strcmp0 (action, "action_decline_4"))
		g_print ("\"No time... I'm riding!\"\n");
	else
		g_print ("That should not have happened (action_decline_4)!");
}

void
action_decline_5 (NotifyNotification* notification,
                  const char*               action,
                  gpointer            data)
{
	if (!g_strcmp0 (action, "action_decline_5"))
		g_print ("Send SMS...\n");
	else
		g_print ("That should not have happened (action_decline_5)!");
}

void
action_accept (NotifyNotification* notification,
               const char*               action,
               gpointer            data)
{
	if (!g_strcmp0 (action, "action_accept"))
		g_print ("Pick up\n");
	else
		g_print ("That should not have happened (action_accept)!");
}

NotifyNotification*
push_notification (const gchar* title,
		   const gchar* body,
		   const gchar* icon)
{
	NotifyNotification* notification;
	gboolean            success;
	GError*             error = NULL;

	/* initial notification */
	notification = notify_notification_new (title, body, icon);

	/* indicate to notification-daemon, that we want to use snap-decisions */
	notify_notification_set_hint_string (notification,
	                                     "x-canonical-snap-decisions",
	                                     "true");

	/* NOTE: the order in which actions are added is important... positive
	** always comes first! */
	notify_notification_add_action (notification,
	                                "action_accept",
	                                "Pick up",
	                                (NotifyActionCallback) action_accept,
	                                NULL,
	                                NULL);

	notify_notification_add_action (notification,
	                                "action_decline_1",
	                                "Hang up",
	                                (NotifyActionCallback) action_decline_1,
	                                NULL,
	                                NULL);

	notify_notification_add_action (notification,
	                                "action_decline_2",
	                                "\"Dude, my wife is here!\"",
	                                (NotifyActionCallback) action_decline_2,
	                                NULL,
	                                NULL);

	notify_notification_add_action (notification,
	                                "action_decline_3",
	                                "\"I'm sleeping.\"",
	                                (NotifyActionCallback) action_decline_3,
	                                NULL,
	                                NULL);

	notify_notification_add_action (notification,
	                                "action_decline_4",
	                                "\"No time... I'm riding!\"",
	                                (NotifyActionCallback) action_decline_4,
	                                NULL,
	                                NULL);

	notify_notification_add_action (notification,
	                                "action_decline_5",
	                                "Send SMS...",
	                                (NotifyActionCallback) action_decline_5,
	                                NULL,
	                                NULL);

	/* set the button-tint hint so that the right/positive button is tinted and
	** not using the stock clear-color */
	notify_notification_set_hint_string (notification,
	                                     "x-canonical-private-button-tint",
	                                     "true");

	error = NULL;
	success = notify_notification_show (notification, &error);
	if (!success)
	{
		g_print ("That did not work ... \"%s\".\n",
			 error->message);
		g_error_free (error);
		return NULL;
	}

	return notification;
}

gboolean
hung_up_timeout (gpointer data)
{
	NotifyNotification* n       = NULL;
	GError*             error   = NULL;
	gboolean            success = FALSE;

	if (data)
	{
		n = (NotifyNotification*)  data;
		success = notify_notification_close (n, &error);
		if (!success)
		{
			g_print ("Closing the notification did not work...... \"%s\".\n",
			         error->message);
			g_error_free (error);
		}
		else
			g_print ("Application is hanging up now");
	}

	return FALSE;
}

int 
main (int    argc,
      char** argv)
{
	/* idenify yourself correctly to libnotify/NotifyOSD, otherwise your
	** notifications will not get through */
	if (!notify_init ("sd-example-incoming-call-canceled"))
		return 1;

	/* call this so we can savely use has_cap(CAP_SOMETHING) later */
	init_caps ();

	/* show what's supported */
	print_caps ();

	/* be nice and check for required capabilities */
	if (!has_cap (CAP_SNAP_DECISIONS) && !has_cap (CAP_BUTTON_TINT))
		return 2;

	NotifyNotification* notification;
	notification = push_notification ("Incoming call",
	                                  "Bro Coly\n+1 (0)555-27439",
	                                  "notification-unknown-call");
	if (!notification)
		return 3;

	/* create the main loop, connect the closed-handler */
	GMainLoop* loop = g_main_loop_new (NULL, FALSE);
	g_signal_connect (G_OBJECT (notification),
	                  "closed",
	                  G_CALLBACK (closed_handler),
	                  loop);

	g_timeout_add (3000, hung_up_timeout, notification);

	/* let it run */
	g_main_loop_run (loop);

	notify_uninit ();

	return 0;
}
