////////////////////////////////////////////////////////////////////////////////
//3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
//      10        20        30        40        50        60        70        80
//
// notify-osd
//
// bubble.c - implements all the rendering of a notification bubble
//
// Copyright 2009 Canonical Ltd.
//
// Authors:
//    Mirco "MacSlow" Mueller <mirco.mueller@canonical.com>
//    David Barth <david.barth@canonical.com>
//
// Contributor(s):
//    Frederic "fredp" Peters <fpeters@gnome.org> (icon-only fix, rev. 204)
//    Eitan Isaacson <eitan@ascender.com> (ATK interface for a11y, rev. 351)
//
// self program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// self program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with self program.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <X11/Xatom.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <pixman.h>
#include <math.h>

#include "bubble.h"
#include "defaults.h"
#include "stack.h"
#include "dbus.h"
#include "util.h"
#include "bubble-window.h"
#include "raico-blur.h"
#include "tile.h"
#include "units.h"

G_DEFINE_TYPE (Bubble, bubble, G_TYPE_OBJECT);

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), BUBBLE_TYPE, BubblePrivate))

struct _BubblePrivate {
	BubbleLayout     layout;
	GtkWidget*       widget;
	gint 		 width;
	gint		 height;
	gboolean         visible;
	guint            timer_id;
	gboolean         mouse_over;
	gfloat           distance;
	gchar*           synchronous;
	gboolean         composited;
	guint            draw_handler_id;
	guint            pointer_update_id;
	tile_t*          tile_background_part;
	tile_t*          tile_background;
	tile_t*          tile_icon;
	tile_t*          tile_secondary_icon;
	tile_t*          tile_title;
	tile_t*          tile_body;
	tile_t*          tile_indicator;
	tile_t*          tile_sd_buttons;
	gint             title_width;
	gint             title_height;
	gint             body_width;
	gint             body_height;
	gboolean         append;
	gboolean         icon_only;
	gint             future_height;
	gboolean         prevent_fade;

	// these will be replaced by class Notification later on
	GString*         title;
	GString*         message_body;
	gboolean         title_needs_refresh;
	gboolean         message_body_needs_refresh;
	guint            id;
	GdkPixbuf*       icon_pixbuf;
	GdkPixbuf*		 sd_answer_pixbuf;
	GdkPixbuf*		 sd_ignore_pixbuf;
	gint             value; // "empty": -2, valid range: -1..101, -1/101 trigger "over/undershoot"-effect
	GString*         sender;
	guint            timeout;
	guint            urgency;
	GdkPixbuf*       secondary_icon_pixbuf;
	//notification_t* notification;

	// used to prevent unneeded updates of the tile-cache, for append-,
	// update or replace-cases, needs to move into class Notification
	GString*         old_icon_filename;
	Notification*	 n;
	cairo_surface_t* scratch;
    Units*           _units;
};

enum
{
	TIMED_OUT,
	VALUE_CHANGED,
	MESSAGE_BODY_DELETED,
	MESSAGE_BODY_INSERTED,
	LAST_SIGNAL
};

enum
{
	R = 0,
	G,
	B,
	A
};

#define TEMPORARY_ICON_PREFIX_WORKAROUND 1
#ifdef TEMPORARY_ICON_PREFIX_WORKAROUND
#warning "--== Using the icon-name-substitution! This is a temp. workaround not going to be maintained for long! ==--"
#define NOTIFY_OSD_ICON_PREFIX "notification"
#endif

// FIXME: self is in class Defaults already, but not yet hooked up so for the
// moment we use the macros here, these values reflect the visual-guideline
// for jaunty notifications
#define TEXT_TITLE_COLOR_R 1.0f
#define TEXT_TITLE_COLOR_G 1.0f
#define TEXT_TITLE_COLOR_B 1.0f
#define TEXT_TITLE_COLOR_A 1.0f

#define TEXT_BODY_COLOR_R  0.91f
#define TEXT_BODY_COLOR_G  0.91f
#define TEXT_BODY_COLOR_B  0.91f
#define TEXT_BODY_COLOR_A  1.0f

#define TEXT_SHADOW_COLOR_R 0.0f
#define TEXT_SHADOW_COLOR_G 0.0f 
#define TEXT_SHADOW_COLOR_B 0.0f 
#define TEXT_SHADOW_COLOR_A 1.0f 

#define BUBBLE_BG_COLOR_R  0.0f
#define BUBBLE_BG_COLOR_G  0.0f
#define BUBBLE_BG_COLOR_B  0.0f
#define BUBBLE_BG_COLOR_A  0.85f

#define INDICATOR_UNLIT_R  1.0f
#define INDICATOR_UNLIT_G  1.0f
#define INDICATOR_UNLIT_B  1.0f
#define INDICATOR_UNLIT_A  0.3f

#define INDICATOR_LIT_R    1.0f
#define INDICATOR_LIT_G    1.0f
#define INDICATOR_LIT_B    1.0f
#define INDICATOR_LIT_A    1.0f

#define FPS                 60
#define PROXIMITY_THRESHOLD 100
#define WINDOW_MIN_OPACITY  0.4f
#define WINDOW_MAX_OPACITY  1.0f

// text drop-shadow should _never_ be bigger than content blur-radius!!!
#define BUBBLE_CONTENT_BLUR_RADIUS 4
#define TEXT_DROP_SHADOW_SIZE      2

//-- private functions ---------------------------------------------------------

static guint g_bubble_signals[LAST_SIGNAL] = { 0 };
gint         g_pointer[2];

void
draw_curved_rect (cairo_t* cr,
				  gdouble  x,
				  gdouble  y,
				  gdouble  width,
				  gdouble  height,
				  gdouble  radius,
				  gboolean aligned)
{
	/* NOTE: values are taken from a manually converted SVG-image of the
	** button-outline provided by the Design-team */

    /*
    ** knot- and control-points:
    **
    **            a  A                       H  t
    **            *--*-----------------------*--*
    **         *b                                  *s
    **        /                                     \
    **       *c                                      *r
    **      /                                         \
    **     *d                                          *q
    **   *e                                              *p
    **   |                                               |
    **   |                                               |
    **   *B                                              *G
    **   |                                               |
    **   |                                               |
    **   *C                                              *F
    **   |                                               |
    **   |                                               |
    **   *f                                              *o
    **     *g                                          *n
    **      \                                         /
    **       *h                                      *m
    **        \                                     /
    **         *i                                  *l
    **            *--*-----------------------*--*
    **            j  D                       E  k
    */

	gdouble corner_width1 = 11.09;
	gdouble corner_width2 = 11.094;
	gdouble corner_height = 12.0;
	gdouble factor        =  1.0;

	/* determine scale-factor of corner-shape */
	if (radius != 12.0)
		factor = 1.0 / 12.0 * radius;

	corner_width1 *= factor;
	corner_width2 *= factor;
	corner_height *= factor;

	// top-left: 11,09/12,0
	gdouble a[2] = {-3.7, 0.0};       // relative values from the SVG-image
	gdouble b[2] = {-6.507, 0.912};     // relative values from the SVG-image
	gdouble c[2] = {-8.279, 2.735};     // relative values from the SVG-image
	gdouble d[2] = {-1.773, 1.822};     // relative values from the SVG-image
	gdouble e[2] = {-2.708, 4.911};     // relative values from the SVG-image

	// bottom-left: 11,094/12,0
	gdouble f[2] = {0.103, 4.355};      // relative values from the SVG-image
	gdouble g[2] = {1.037, 7.444};      // relative values from the SVG-image
	gdouble h[2] = {2.811, 9.267};      // relative values from the SVG-image
	gdouble i[2] = {1.772, 1.823};      // relative values from the SVG-image
	gdouble j[2] = {4.531, 2.735};      // relative values from the SVG-image

	// bottom-right: 11,09/12,0
	gdouble k[2] = {3.748,  0.0};       // relative values from the SVG-image
	gdouble l[2] = {6.507, -0.912};     // relative values from the SVG-image
	gdouble m[2] = {8.279, -2.735};     // relative values from the SVG-image
	gdouble n[2] = {1.773, -1.822};     // relative values from the SVG-image
	gdouble o[2] = {2.708, -4.911};     // relative values from the SVG-image

	// top-right: 11,094/12,0
	gdouble p[2] = {-0.103, -4.355};    // relative values from the SVG-image
	gdouble q[2] = {-1.037, -7.444};    // relative values from the SVG-image
	gdouble r[2] = {-2.811, -9.267};    // relative values from the SVG-image
	gdouble s[2] = {-1.772, -1.823};    // relative values from the SVG-image
	gdouble t[2] = {-4.531, -2.735};    // relative values from the SVG-image

	gdouble A[2] = {x + corner_width1, y};
	gdouble B[2] = {x, y + corner_height};
	gdouble C[2] = {x, y + height - corner_height};
	gdouble D[2] = {x + corner_width2, y + height};
	gdouble E[2] = {x + width - corner_width1, y + height};
	gdouble F[2] = {x + width, y + height - corner_height};
	gdouble G[2] = {x + width, y + corner_height};
	gdouble H[2] = {x + width - corner_width2, y};

	a[0] *= factor;
	a[1] *= factor;
	b[0] *= factor;
	b[1] *= factor;
	c[0] *= factor;
	c[1] *= factor;
	d[0] *= factor;
	d[1] *= factor;
	e[0] *= factor;
	e[1] *= factor;

	f[0] *= factor;
	f[1] *= factor;
	g[0] *= factor;
	g[1] *= factor;
	h[0] *= factor;
	h[1] *= factor;
	i[0] *= factor;
	i[1] *= factor;
	j[0] *= factor;
	j[1] *= factor;

	k[0] *= factor;
	k[1] *= factor;
	l[0] *= factor;
	l[1] *= factor;
	m[0] *= factor;
	m[1] *= factor;
	n[0] *= factor;
	n[1] *= factor;
	o[0] *= factor;
	o[1] *= factor;

	p[0] *= factor;
	p[1] *= factor;
	q[0] *= factor;
	q[1] *= factor;
	r[0] *= factor;
	r[1] *= factor;
	s[0] *= factor;
	s[1] *= factor;
	t[0] *= factor;
	t[1] *= factor;

	/* top-left corner */
	cairo_move_to (cr, A[0], align (A[1], aligned));
	cairo_curve_to (cr,
					A[0] + a[0], A[1] + a[1],
					A[0] + b[0], A[1] + b[1],
					A[0] + c[0], A[1] + c[1]);
	cairo_curve_to (cr,
					A[0] + c[0] + d[0], A[1] + c[1] + d[1],
					A[0] + c[0] + e[0], A[1] + c[1] + e[1],
					align (B[0], aligned), B[1]);

	/* top-left-to-bottom-left straight */
	cairo_line_to (cr, align (C[0], aligned), C[1]);

	/* bottom-left corner */
	cairo_curve_to (cr,
					C[0] + f[0], C[1] + f[1],
					C[0] + g[0], C[1] + g[1],
					C[0] + h[0], C[1] + h[1]);
	cairo_curve_to (cr,
					C[0] + h[0] + i[0], C[1] + h[1] + i[1],
					C[0] + h[0] + j[0], C[1] + h[1] + j[1],
					D[0], align (D[1], aligned));

	/* bottom-left-to-right straight */
	cairo_line_to (cr, E[0], align (E[1], aligned));

	/* bottom-right corner */
	cairo_curve_to (cr,
					E[0] + k[0], E[1] + k[1],
					E[0] + l[0], E[1] + l[1],
					E[0] + m[0], E[1] + m[1]);
	cairo_curve_to (cr,
					E[0] + m[0] + n[0], E[1] + m[1] + n[1],
					E[0] + m[0] + o[0], E[1] + m[1] + o[1],
					align (F[0], aligned), F[1]);

	/* right-bottom-to-top straight */
	cairo_line_to (cr, align (G[0], aligned), G[1]);

	/* top-right corner */
	cairo_curve_to (cr,
					G[0] + p[0], G[1] + p[1],
					G[0] + q[0], G[1] + q[1],
					G[0] + r[0], G[1] + r[1]);
	cairo_curve_to (cr,
					G[0] + r[0] + s[0], G[1] + r[1] + s[1],
					G[0] + r[0] + t[0], G[1] + r[1] + t[1],
					H[0], align (H[1], aligned));

	/* top-right-to-left straight */
	cairo_close_path (cr);
}

void
draw_round_rect (cairo_t* cr,
				 gdouble  aspect,        // aspect-ratio
				 gdouble  x,             // top-left corner
				 gdouble  y,             // top-left corner
				 gdouble  corner_radius, // "size" of the corners
				 gdouble  width,         // width of the rectangle
				 gdouble  height)        // height of the rectangle
{
	gdouble radius = corner_radius / aspect;

	// top-left, right of the corner
	cairo_move_to (cr, x + radius, y);

	// top-right, left of the corner
	cairo_line_to (cr,
		       x + width - radius,
		       y);

	// top-right, below the corner
	cairo_arc (cr,
                   x + width - radius,
                   y + radius,
                   radius,
                   -90.0f * G_PI / 180.0f,
                   0.0f * G_PI / 180.0f);

	// bottom-right, above the corner
	cairo_line_to (cr,
		       x + width,
		       y + height - radius);

	// bottom-right, left of the corner
	cairo_arc (cr,
                   x + width - radius,
                   y + height - radius,
                   radius,
                   0.0f * G_PI / 180.0f,
                   90.0f * G_PI / 180.0f);

	// bottom-left, right of the corner
	cairo_line_to (cr,
		       x + radius,
		       y + height);

	// bottom-left, above the corner
	cairo_arc (cr,
                   x + radius,
                   y + height - radius,
                   radius,
                   90.0f * G_PI / 180.0f,
                   180.0f * G_PI / 180.0f);

	// top-left, below the corner
	cairo_line_to (cr,
		       x,
		       y + radius);

	// top-left, right of the corner
	cairo_arc (cr,
                   x + radius,
                   y + radius,
                   radius,
                   180.0f * G_PI / 180.0f,
                   270.0f * G_PI / 180.0f);
}

void
draw_button_label (cairo_t*     cr,
                   const gchar* label,
                   gint         x,
                   gint         y,
                   gint         width,
                   gint         height,
                   gint         horiz_margin,
                   gchar* font_face,
                   gdouble      font_size, /* in pixels */
                   PangoWeight  font_weight,
                   gdouble*     font_color)
{
	if (width == 0 || cr == NULL || cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return;
  Defaults*             d = defaults_get_default();
	PangoFontDescription* desc     = NULL;
	PangoLayout*          layout   = NULL;
	PangoRectangle        log_rect = {0, 0, 0, 0};

	// create pango desc/layout
	layout = pango_cairo_create_layout (cr);
	desc = pango_font_description_from_string (font_face);

	pango_font_description_set_absolute_size (desc, font_size * PANGO_SCALE);
	pango_font_description_set_weight (desc, font_weight);
	pango_layout_set_font_description (layout, desc);
	pango_font_description_free (desc);
	pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_NONE);
	/* just make it wide enought so it'll fit in a single line */
	pango_layout_set_width (layout, 1000 * PANGO_SCALE);
	pango_layout_set_height (layout, -1);

	// print and layout string (pango-wise)
	pango_layout_set_text (layout, label, -1);
	pango_layout_get_extents (layout, NULL, &log_rect);

	if ((width - 2 * horiz_margin) < (log_rect.width / PANGO_SCALE))
	{
		gdouble factor = 1.0;
		gdouble tw     = (gdouble) log_rect.width / (gdouble) PANGO_SCALE;
		gdouble w      = (gdouble) width - 2.0 * (gdouble) horiz_margin;

		factor = w / tw;
		pango_font_description_set_absolute_size ((PangoFontDescription*)pango_layout_get_font_description (layout),
		                                          factor * font_size * PANGO_SCALE);
		pango_layout_set_width (layout, (width - 2 * horiz_margin) * PANGO_SCALE);
		pango_layout_get_extents (layout, NULL, &log_rect);
	}

	// make sure system-wide font-options like hinting, antialiasing etc.
	// are taken into account
	pango_cairo_context_set_resolution  (pango_layout_get_context (layout),
					     defaults_get_screen_dpi (d));
	pango_layout_context_changed (layout);

	// now draw text
	gdouble offset_x = x + (width - log_rect.width / PANGO_SCALE) / 2.0;
	gdouble offset_y = y + (height - log_rect.height / PANGO_SCALE) / 2.0;
	cairo_move_to (cr, offset_x, offset_y);
	cairo_set_source_rgba (cr,
	                       font_color[0],
	                       font_color[1],
	                       font_color[2],
	                       font_color[3]);
	pango_cairo_show_layout (cr, layout);

	g_object_unref (layout);
}

void
draw_text (cairo_t*     cr,
               const gchar* label,
               gint         x,
               gint         y,
               gint         width,
               gint         height,
               gint         line_limit,
               gchar*       font_face,
               gdouble      font_size,
               PangoWeight  font_weight,
               gdouble*     font_color)
{
	if (width == 0 || cr == NULL || cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return;

  Defaults*             d = defaults_get_default();
	PangoFontDescription* desc     = NULL;
	PangoLayout*          layout   = NULL;
	PangoRectangle        log_rect = {0, 0, 0, 0};

	// create pango desc/layout
	layout = pango_cairo_create_layout (cr);
	desc = pango_font_description_from_string (font_face);

	pango_font_description_set_absolute_size (desc, font_size * PANGO_SCALE);
	pango_font_description_set_weight (desc, font_weight);
	pango_layout_set_font_description (layout, desc);
	pango_font_description_free (desc);
	pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
	/* just make it wide enought so it'll fit in a single line */
	pango_layout_set_width (layout, width * PANGO_SCALE);
	pango_layout_set_height (layout, -line_limit);

	// print and layout string (pango-wise)
	pango_layout_set_text (layout, label, -1);
	pango_layout_get_extents (layout, NULL, &log_rect);

	if (line_limit == 1)
	{
		if (width < (log_rect.width / PANGO_SCALE))
		{
			gdouble factor = 1.0;
			gdouble delta  = 0.0;
			gdouble tw     = log_rect.width / (gdouble) PANGO_SCALE;
			gdouble w      = (gdouble) width;

			delta  = tw - w;
			factor = abs ((delta * 100.0 / w) - 100.0) / 100.0;
			pango_font_description_set_absolute_size ((PangoFontDescription*) pango_layout_get_font_description (layout),
				                                      factor * font_size * PANGO_SCALE);
			pango_layout_set_width (layout, width * PANGO_SCALE);
			pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_NONE);
			pango_layout_get_extents (layout, NULL, &log_rect);
		}
	}

	// make sure system-wide font-options like hinting, antialiasing etc.
	// are taken into account
	pango_cairo_context_set_resolution  (pango_layout_get_context (layout),
					     defaults_get_screen_dpi (d));
	pango_layout_context_changed (layout);

	// now draw text
	gdouble offset_x = x;
	gdouble offset_y = y;
	cairo_move_to (cr, offset_x, offset_y);
	cairo_set_source_rgba (cr,
	                       font_color[0],
	                       font_color[1],
	                       font_color[2],
	                       font_color[3]);
	pango_cairo_show_layout (cr, layout);

	g_object_unref (layout);
}

gint
get_text_height (gchar*       label,
                 gint         max_width,
                 gint         line_limit,
                 gchar*       font_face,
                 gdouble      font_size,
                 PangoWeight  font_weight)
{
	if (!label)
		return 0;

	cairo_surface_t*      surface  = NULL;
	cairo_t*              cr       = NULL;
	PangoFontDescription* desc     = NULL;
	PangoLayout*          layout   = NULL;
	PangoRectangle        log_rect = {0, 0, 0, 0};

	// create pango desc/layout from dummy surface
	surface = cairo_image_surface_create (CAIRO_FORMAT_A1, 1, 1);
	cr      = cairo_create (surface);
	layout  = pango_cairo_create_layout (cr);
	desc    = pango_font_description_from_string (font_face);

	cairo_surface_destroy (surface);
	cairo_destroy (cr);

	pango_font_description_set_absolute_size (desc, font_size * PANGO_SCALE);
	pango_font_description_set_weight (desc, font_weight);
	pango_layout_set_font_description (layout, desc);
	pango_font_description_free (desc);
	pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
	/* just make it wide enought so it'll fit in a single line */
	pango_layout_set_width (layout, max_width * PANGO_SCALE);
	pango_layout_set_height (layout, -line_limit);

	// print and layout string (pango-wise)
	pango_layout_set_text (layout, label, -1);
	pango_layout_get_extents (layout, NULL, &log_rect);

	g_object_unref (layout);

	return log_rect.height / PANGO_SCALE;
}

// color-, alpha-, radius-, width-, height- and gradient-values were determined
// by very close obvervation of a SVG-mockup from the design-team
static void
_draw_value_indicator (cairo_t* cr,
		       gint     value,             // value to render: 0 - 100
		       gint     start_x,           // top of surrounding rect
		       gint     start_y,           // left of surrounding rect
		       gint     width,             // width of surrounding rect
		       gint     height,            // height of surrounding rect
		       gint     outline_thickness) // outline-thickness
{
	gdouble          outline_radius;
	gdouble          outline_width;
	gdouble          outline_height;
	gdouble          bar_radius;
	gdouble          bar_width;
	//gdouble          bar_height;
	cairo_pattern_t* gradient;

	outline_radius = outline_thickness;
	outline_width  = width;
	outline_height = height;

	// draw bar-background
	cairo_set_line_width (cr, outline_thickness);
	cairo_set_source_rgba (cr, 0.0f, 0.0f, 0.0f, 0.5f);
	draw_round_rect (cr,
			 1.0f,
			 start_x,
			 start_y,
			 outline_radius,
			 outline_width,
			 outline_height);
	cairo_fill (cr);
	gradient = cairo_pattern_create_linear (0.0f,
						start_y + outline_thickness,
						0.0f,
						start_y +
						outline_height -
	                                        2 * outline_thickness);
	cairo_pattern_add_color_stop_rgba (gradient,
					   0.0f,
					   0.866f,
					   0.866f,
					   0.866f,
					   0.3f);
	cairo_pattern_add_color_stop_rgba (gradient,
					   0.2f,
					   0.827f,
					   0.827f,
					   0.827f,
					   0.3f);
	cairo_pattern_add_color_stop_rgba (gradient,
					   0.3f,
					   0.772f,
					   0.772f,
					   0.772f,
					   0.3f);
	cairo_pattern_add_color_stop_rgba (gradient,
					   1.0f,
					   0.623f,
					   0.623f,
					   0.623f,
					   0.3f);
	cairo_set_source (cr, gradient);
	draw_round_rect (cr,
			 1.0f,
			 start_x + outline_thickness,
			 start_y + outline_thickness,
			 outline_radius,
			 outline_width - 2 * outline_thickness,
			 outline_height - 2 * outline_thickness);
	cairo_fill (cr);
	cairo_pattern_destroy (gradient);

	bar_radius = outline_radius;
	bar_width  = outline_width - 2 * outline_radius;
	//bar_height = outline_height - outline_radius;

	// draw value-bar
	if (value > 0)
	{
		gint corrected_value = value;

		if (corrected_value > 100)
			corrected_value = 100;

		draw_round_rect (cr,
				 1.0f,
				 start_x + outline_thickness,
				 start_y + outline_thickness,
				 bar_radius,
				 bar_width / 100.0f * (gdouble) corrected_value,
				 outline_height - 2 * outline_thickness);

		gradient = cairo_pattern_create_linear (0.0f,
							start_y +
							outline_thickness,
		                                        0.0f,
							start_y +
							outline_height -
							2 * outline_thickness);
		cairo_pattern_add_color_stop_rgba (gradient,
						   0.0f,
						   0.9f,
						   0.9f,
						   0.9f,
						   1.0f);
		cairo_pattern_add_color_stop_rgba (gradient,
						   0.75f,
						   0.5f,
						   0.5f,
						   0.5f,
						   1.0f);
		cairo_pattern_add_color_stop_rgba (gradient,
						   1.0f,
						   0.4f,
						   0.4f,
						   0.4f,
						   1.0f);

		cairo_set_source (cr, gradient);
		cairo_fill (cr);
		cairo_pattern_destroy (gradient);
	}
}

void
_draw_shadow (cairo_t* cr,
	      gdouble  width,
	      gdouble  height,
	      gint     shadow_radius,
	      gint     corner_radius)
{
	cairo_surface_t* tmp_surface     = NULL;
	cairo_surface_t* new_surface     = NULL;
	cairo_pattern_t* pattern         = NULL;
	cairo_t*         cr_surf         = NULL;
	cairo_matrix_t   matrix;
	raico_blur_t*    blur            = NULL;

	tmp_surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
						  4 * shadow_radius,
						  4 * shadow_radius);
	if (cairo_surface_status (tmp_surface) != CAIRO_STATUS_SUCCESS) {
		if (tmp_surface)
			cairo_surface_destroy (tmp_surface);
		return;
	}

	cr_surf = cairo_create (tmp_surface);
	if (cairo_status (cr_surf) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (tmp_surface);
		if (cr_surf)
			cairo_destroy (cr_surf);
		return;
	}

	cairo_scale (cr_surf, 1.0f, 1.0f);
	cairo_set_operator (cr_surf, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr_surf);
	cairo_set_operator (cr_surf, CAIRO_OPERATOR_OVER);
	cairo_set_source_rgba (cr_surf, 0.0f, 0.0f, 0.0f, 1.0f);
	cairo_arc (cr_surf,
		   2 * shadow_radius,
		   2 * shadow_radius,
		   1.75f * corner_radius,
		   0.0f,
		   360.0f * (G_PI / 180.f));
	cairo_fill (cr_surf);
	cairo_destroy (cr_surf);

	// create and setup blur
	blur = raico_blur_create (RAICO_BLUR_QUALITY_HIGH);
	raico_blur_set_radius (blur, shadow_radius);

	// now blur it
	raico_blur_apply (blur, tmp_surface);

	// blur no longer needed
	raico_blur_destroy (blur);

	new_surface = cairo_image_surface_create_for_data (
			cairo_image_surface_get_data (tmp_surface),
			cairo_image_surface_get_format (tmp_surface),
			cairo_image_surface_get_width (tmp_surface) / 2,
			cairo_image_surface_get_height (tmp_surface) / 2,
			cairo_image_surface_get_stride (tmp_surface));
	pattern = cairo_pattern_create_for_surface (new_surface);
	if (cairo_pattern_status (pattern) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (tmp_surface);
		cairo_surface_destroy (new_surface);
		if (pattern)
			cairo_pattern_destroy (pattern);
		return;
	}

	// top left
	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_PAD);
	cairo_set_source (cr, pattern);
	cairo_rectangle (cr,
			 0.0f,
			 0.0f,
			 width - 2 * shadow_radius,
			 2 * shadow_radius);
	cairo_fill (cr);

	// bottom left
	cairo_matrix_init_scale (&matrix, 1.0f, -1.0f);
	cairo_matrix_translate (&matrix, 0.0f, -height);
	cairo_pattern_set_matrix (pattern, &matrix);
	cairo_rectangle (cr,
			 0.0f,
			 2 * shadow_radius,
			 2 * shadow_radius,
			 height - 2 * shadow_radius);
	cairo_fill (cr);

	// top right
	cairo_matrix_init_scale (&matrix, -1.0f, 1.0f);
	cairo_matrix_translate (&matrix, -width, 0.0f);
	cairo_pattern_set_matrix (pattern, &matrix);
	cairo_rectangle (cr,
			 width - 2 * shadow_radius,
			 0.0f,
			 2 * shadow_radius,
			 height - 2 * shadow_radius);
	cairo_fill (cr);

	// bottom right
	cairo_matrix_init_scale (&matrix, -1.0f, -1.0f);
	cairo_matrix_translate (&matrix, -width, -height);
	cairo_pattern_set_matrix (pattern, &matrix);
	cairo_rectangle (cr,
			 2 * shadow_radius,
			 height - 2 * shadow_radius,
			 width - 2 * shadow_radius,
			 2 * shadow_radius);
	cairo_fill (cr);

	// clean up
	cairo_pattern_destroy (pattern);
	cairo_surface_destroy (tmp_surface);
	cairo_surface_destroy (new_surface);
}

static void
_draw_layout_grid (cairo_t* cr,
		  Bubble*  bubble)
{
	Defaults* d = bubble->defaults;

	if (!cr)
		return;

	cairo_set_line_width (cr, 1.0f);
	cairo_set_source_rgba (cr, 1.0f, 0.5f, 0.25f, 1.0f);

	// all vertical grid lines
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d) +
		       EM2PIXELS (defaults_get_icon_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d) +
		       EM2PIXELS (defaults_get_icon_size (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (2 * defaults_get_margin_size (d), d) +
		       EM2PIXELS (defaults_get_icon_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (2 * defaults_get_margin_size (d), d) +
		       EM2PIXELS (defaults_get_icon_size (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d) -
		       EM2PIXELS (defaults_get_margin_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d) -
		       EM2PIXELS (defaults_get_margin_size (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));

	// all horizontal grid lines
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d) +
		       EM2PIXELS (defaults_get_icon_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d),
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_margin_size (d), d) +
		       EM2PIXELS (defaults_get_icon_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d) -
		       EM2PIXELS (defaults_get_margin_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d) -
		       EM2PIXELS (defaults_get_margin_size (d), d));
	cairo_move_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));
	cairo_line_to (cr,
		       0.5f + EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
		       EM2PIXELS (defaults_get_bubble_width (d), d),
		       0.5f + (gdouble) bubble_get_height (bubble) -
		       EM2PIXELS (defaults_get_bubble_shadow_size (d), d));

	cairo_stroke (cr);
}

void
_refresh_background (Bubble* self)
{
	BubblePrivate*   priv       = GET_PRIVATE (self);
	Defaults*        d          = self->defaults;
	cairo_t*         cr         = NULL;
	cairo_surface_t* scratch    = NULL;
	cairo_surface_t* dummy      = NULL;
	cairo_surface_t* clone      = NULL;
	cairo_surface_t* normal     = NULL;
	cairo_surface_t* blurred    = NULL;
	raico_blur_t*    blur       = NULL;
	gint             width;
	gint             height;
    gint             gu         = priv->_units->getGU ();

	bubble_get_size (self, &width, &height);

	// create temp. scratch surface for top-left shadow/background part
	scratch = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 3 * gu, 3 * gu);

	g_return_if_fail (scratch);

	if (cairo_surface_status (scratch) != CAIRO_STATUS_SUCCESS)
	{
		if (scratch)
			cairo_surface_destroy (scratch);

		return;
	}

	// create drawing context for that temp. scratch surface
	cr = cairo_create (scratch);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (scratch);

		if (cr)
			cairo_destroy (cr);

		return;
	}

	// clear, render top-left part of shadow/background in scratch-surface
    cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	/*_draw_shadow (
		cr,
		width,
		height,
		gu,
		gu);*/
	/*cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	draw_curved_rect (cr,
					  gu,
					  gu,
					  38 * gu,
					  bubble_get_height (self) - 2 * gu,
					  gu,
					  FALSE);
	cairo_fill (cr);*/
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_set_source_rgba (cr,
			       BUBBLE_BG_COLOR_R,
			       BUBBLE_BG_COLOR_G,
			       BUBBLE_BG_COLOR_B,
			       BUBBLE_BG_COLOR_A);

	draw_curved_rect (cr,
					  0,
					  gu,
					  38 * gu,
					  bubble_get_height (self),
					  gu,
					  FALSE);
	cairo_fill (cr);
	cairo_destroy (cr);

	// create temp. clone of scratch surface
	dummy = cairo_image_surface_create_for_data (
			cairo_image_surface_get_data (scratch),
			cairo_image_surface_get_format (scratch),
			3 * gu,
			3 * gu,
			cairo_image_surface_get_stride (scratch));
	clone = copy_surface (dummy);
	cairo_surface_destroy (dummy);

	// create normal surface from that surface-clone
	dummy = cairo_image_surface_create_for_data (
			cairo_image_surface_get_data (clone),
			cairo_image_surface_get_format (clone),
			2 * gu,
			2 * gu,
			cairo_image_surface_get_stride (clone));
	normal = copy_surface (dummy);
	cairo_surface_destroy (dummy);

	// now blur the surface-clone
	/*blur = raico_blur_create (RAICO_BLUR_QUALITY_LOW);
	raico_blur_set_radius (blur, BUBBLE_CONTENT_BLUR_RADIUS);
	raico_blur_apply (blur, clone);
	raico_blur_destroy (blur);*/

	// create blurred version from that blurred surface-clone 
	dummy = cairo_image_surface_create_for_data (
			cairo_image_surface_get_data (clone),
			cairo_image_surface_get_format (clone),
			2 * gu,
			2 * gu,
			cairo_image_surface_get_stride (clone));
	blurred = copy_surface (dummy);
	cairo_surface_destroy (dummy);
	destroy_cloned_surface (clone);

	// finally create tile with top-left shadow/background part
	if (priv->tile_background_part)
		tile_destroy (priv->tile_background_part);
	priv->tile_background_part = tile_new_for_padding (normal, blurred);
	destroy_cloned_surface (normal);
	destroy_cloned_surface (blurred);

	// create surface(s) for full shadow/background tile
	// we need two RGBA-surfaces
	normal = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
					     width,
					     height);
	if (cairo_surface_status (normal) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (scratch);

		if (normal)
			cairo_surface_destroy (normal);

		return;
	}

	blurred = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
					      width,
					      height);
	if (cairo_surface_status (blurred) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		cairo_surface_destroy (scratch);

		if (blurred)
			cairo_surface_destroy (blurred);

		return;
	}

	// use tile for top-left background-part to fill the full bg-surface
	// create context for blurred surface
	cr = cairo_create (blurred);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		cairo_surface_destroy (blurred);
		cairo_surface_destroy (scratch);

		if (cr)
			cairo_destroy (cr);

		return;
	}

	// clear blurred surface
    cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	// fill with blurred state of background-part tile
	tile_paint_with_padding (priv->tile_background_part,
				 cr,
				 0.0f,
				 0.0f,
				 width,
				 height,
				 0.0f,
				 1.0f);

	// get rid of context for blurred surface
	cairo_destroy (cr);

	// create context for normal surface
	cr = cairo_create (normal);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		cairo_surface_destroy (scratch);

		if (priv->composited)
			cairo_surface_destroy (blurred);

		return;
	}

	// clear normal surface
    cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	// fill with normal state of background-part tile
	tile_paint_with_padding (priv->tile_background_part,
				 cr,
				 0.0f,
				 0.0f,
				 width,
				 height,
				 1.0f,
				 0.0f);

	// get rid of context for normal surface
	cairo_destroy (cr);

	// finally create tile for full background
	if (priv->tile_background)
		tile_destroy (priv->tile_background);
	if (priv->composited)
		priv->tile_background = tile_new_for_padding (normal, blurred);
	else
		priv->tile_background = tile_new_for_padding (normal, normal);

	cairo_surface_destroy (blurred);

	cairo_surface_destroy (normal);
	cairo_surface_destroy (scratch);
}

void
_refresh_sd_buttons (Bubble* self)
{
	BubblePrivate*   priv   = GET_PRIVATE (self);
	Defaults*        d      = self->defaults;
	cairo_surface_t* normal = NULL;
	cairo_t*         cr     = NULL;
    gint             gu     = priv->_units->getGU ();

	if (!priv->sd_answer_pixbuf || !priv->sd_ignore_pixbuf)
		return;

	// create temp. scratch surface
	normal = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 36 * gu, 4 * gu);
	if (cairo_surface_status (normal) != CAIRO_STATUS_SUCCESS)
		return;

	// create context for normal surface
	cr = cairo_create (normal);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		return;
	}

	// clear normal surface
	cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	// render buttons into normal surface
	gdk_cairo_set_source_pixbuf (cr,
								 priv->sd_ignore_pixbuf,
								 0,
								 0);
	cairo_paint (cr);
	gdk_cairo_set_source_pixbuf (cr,
								 priv->sd_answer_pixbuf,
								 19 * gu,
								 0);
	cairo_paint (cr);

	// create the surface/blur-cache from the normal surface
	if (priv->tile_sd_buttons)
		tile_destroy (priv->tile_sd_buttons);
	priv->tile_sd_buttons = tile_new (normal, 1);

	// clean up
	cairo_destroy (cr);
	cairo_surface_destroy (normal);
}

void
_refresh_icon (Bubble* self)
{
	BubblePrivate*   priv   = GET_PRIVATE (self);
	Defaults*        d      = self->defaults;
	cairo_surface_t* normal = NULL;
	cairo_t*         cr     = NULL;
    gint             gu     = priv->_units->getGU ();

	if (!priv->icon_pixbuf)
		return;

	// create temp. scratch surface
	if (bubble_get_layout (self) == LAYOUT_ICON_TITLE)
		normal = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 2 * gu, 2 * gu);
	else
		normal = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 6 * gu, 6 * gu);
	
	if (cairo_surface_status (normal) != CAIRO_STATUS_SUCCESS)
		return;

	// create context for normal surface
	cr = cairo_create (normal);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		return;
	}

	// clear normal surface
    cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	// render icon into normal surface
	gdk_cairo_set_source_pixbuf (cr, priv->icon_pixbuf, 0, 0);
	cairo_paint (cr);

	// create the surface/blur-cache from the normal surface
	if (priv->tile_icon)
		tile_destroy (priv->tile_icon);
	priv->tile_icon = tile_new (normal, BUBBLE_CONTENT_BLUR_RADIUS/2);

	// clean up
	cairo_destroy (cr);
	cairo_surface_destroy (normal);
}

void
_refresh_secondary_icon (Bubble* self)
{
	BubblePrivate*   priv   = GET_PRIVATE (self);
	Defaults*        d      = self->defaults;
	cairo_surface_t* normal = NULL;
	cairo_t*         cr     = NULL;
    gint             gu     = priv->_units->getGU ();

	if (!priv->secondary_icon_pixbuf)
		return;

	// create temp. scratch surface
	normal = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 2 * gu, 2 * gu);
	if (cairo_surface_status (normal) != CAIRO_STATUS_SUCCESS)
		return;

	// create context for normal surface
	cr = cairo_create (normal);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		return;
	}

	// clear normal surface
	cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	// render icon into normal surface
	gdk_cairo_set_source_pixbuf (cr, priv->secondary_icon_pixbuf, 0.0, 0.0);
	cairo_paint (cr);

	// create the surface/blur-cache from the normal surface
	if (priv->tile_secondary_icon)
		tile_destroy (priv->tile_secondary_icon);
	priv->tile_secondary_icon = tile_new (normal, 1);

	// clean up
	cairo_destroy (cr);
	cairo_surface_destroy (normal);
}

void
_refresh_title (Bubble* self)
{
	BubblePrivate*        priv           = GET_PRIVATE (self);
	Defaults*             d              = self->defaults;
	cairo_surface_t*      normal         = NULL;
	cairo_t*              cr             = NULL;
	PangoFontDescription* desc           = NULL;
	PangoLayout*          layout         = NULL;
	raico_blur_t*         blur           = NULL;
	gchar*                text_font_face = NULL;

	// create temp. scratch surface
	normal = cairo_image_surface_create (
			CAIRO_FORMAT_ARGB32,
			priv->title_width + 2 * BUBBLE_CONTENT_BLUR_RADIUS,
			priv->title_height + 2 * BUBBLE_CONTENT_BLUR_RADIUS);
	if (cairo_surface_status (normal) != CAIRO_STATUS_SUCCESS)
		return;

	// create context for normal surface
	cr = cairo_create (normal);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		return;
	}

	// clear normal surface
    	cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	// create pango desc/layout
	layout = pango_cairo_create_layout (cr);
	//text_font_face = defaults_get_text_font_face (d);
	desc = pango_font_description_from_string ("Ubuntu Medium");
	//g_free ((gpointer) text_font_face);

	pango_font_description_set_absolute_size (desc,
											  priv->_units->sizeToPixel ("medium") * PANGO_SCALE);
	/*pango_font_description_set_size (desc,
					 defaults_get_system_font_size (d) *
					 defaults_get_text_title_size (d) *
					 PANGO_SCALE);*/

	/*pango_font_description_set_weight (desc,
					   (PangoWeight)defaults_get_text_title_weight (d));*/
	pango_layout_set_font_description (layout, desc);
	pango_font_description_free (desc);

	pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
	pango_layout_set_width (layout, priv->title_width * PANGO_SCALE);
	pango_layout_set_height (layout, priv->title_height * PANGO_SCALE);

	// print and layout string (pango-wise)
	pango_layout_set_text (layout, priv->title->str, priv->title->len);

	// make sure system-wide font-options like hinting, antialiasing etc.
	// are taken into account
	pango_cairo_context_set_resolution  (pango_layout_get_context (layout),
					     defaults_get_screen_dpi (d));
	pango_layout_context_changed (layout);

	// draw text for drop-shadow and ...
	cairo_move_to (cr, 0, 0);

	/*cairo_set_source_rgba (cr,
			       TEXT_SHADOW_COLOR_R,
			       TEXT_SHADOW_COLOR_G,
			       TEXT_SHADOW_COLOR_B,
			       TEXT_SHADOW_COLOR_A);

	pango_cairo_show_layout (cr, layout);*/

	// ... blur it
	/*blur = raico_blur_create (RAICO_BLUR_QUALITY_LOW);
	raico_blur_set_radius (blur, TEXT_DROP_SHADOW_SIZE);
	raico_blur_apply (blur, normal);
	raico_blur_destroy (blur);*/

	// now draw normal (non-blurred) text over drop-shadow
	cairo_set_source_rgba (cr, 0.952941176, 0.952941176, 0.905882353, 1.0);
	/*cairo_set_source_rgba (cr,
			       TEXT_TITLE_COLOR_R,
			       TEXT_TITLE_COLOR_G,
			       TEXT_TITLE_COLOR_B,
			       TEXT_TITLE_COLOR_A);*/
	pango_cairo_show_layout (cr, layout);

	g_object_unref (layout);

	// create the surface/blur-cache from the normal surface
	if (priv->tile_title)
		tile_destroy (priv->tile_title);
	priv->tile_title = tile_new (normal, BUBBLE_CONTENT_BLUR_RADIUS);

	// clean up
	cairo_destroy (cr);
	cairo_surface_destroy (normal);

	// reset refresh-flag
	priv->title_needs_refresh = FALSE;
}

void
_refresh_body (Bubble* self)
{
	BubblePrivate*        priv           = GET_PRIVATE (self);
	Defaults*             d              = self->defaults;
	cairo_surface_t*      normal         = NULL;
	cairo_t*              cr             = NULL;
	PangoFontDescription* desc           = NULL;
	PangoLayout*          layout         = NULL;
	raico_blur_t*         blur           = NULL;
	gchar*                text_font_face = NULL;

	// create temp. scratch surface
	normal = cairo_image_surface_create (
			CAIRO_FORMAT_ARGB32,
			priv->body_width + 2 * BUBBLE_CONTENT_BLUR_RADIUS,
			priv->body_height + 2 * BUBBLE_CONTENT_BLUR_RADIUS);
	if (cairo_surface_status (normal) != CAIRO_STATUS_SUCCESS)
		return;

	// create context for normal surface
	cr = cairo_create (normal);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		return;
	}

	// clear normal surface
    cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	// create pango desc/layout
	layout = pango_cairo_create_layout (cr);
	//text_font_face = defaults_get_text_font_face (d);
	desc = pango_font_description_from_string ("Ubuntu");
	//g_free ((gpointer) text_font_face);

	pango_font_description_set_absolute_size (desc,
											  priv->_units->sizeToPixel ("small") * PANGO_SCALE);
	/*pango_font_description_set_size (desc,
					 defaults_get_system_font_size (d) *
					 defaults_get_text_body_size (d) *
					 PANGO_SCALE);*/

	/*pango_font_description_set_weight (desc,
					   (PangoWeight)defaults_get_text_body_weight (d));*/
	pango_layout_set_font_description (layout, desc);
	pango_font_description_free (desc);

	pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
	pango_layout_set_width (layout, priv->body_width * PANGO_SCALE);
	pango_layout_set_height (layout, priv->body_height * PANGO_SCALE);

	// print and layout string (pango-wise)
	pango_layout_set_text (layout,
	                       priv->message_body->str,
	                       priv->message_body->len);

	// make sure system-wide font-options like hinting, antialiasing etc.
	// are taken into account
	pango_cairo_context_set_resolution  (pango_layout_get_context (layout),
					     defaults_get_screen_dpi (d));
	pango_layout_context_changed (layout);

	// draw text for drop-shadow and ...
	cairo_move_to (cr, 0, 0);

	/*cairo_set_source_rgba (cr,
			       TEXT_SHADOW_COLOR_R,
			       TEXT_SHADOW_COLOR_G,
			       TEXT_SHADOW_COLOR_B,
			       TEXT_SHADOW_COLOR_A);
	pango_cairo_show_layout (cr, layout);*/

	// ... blur it
	/*blur = raico_blur_create (RAICO_BLUR_QUALITY_LOW);
	raico_blur_set_radius (blur, TEXT_DROP_SHADOW_SIZE);
	raico_blur_apply (blur, normal);
	raico_blur_destroy (blur);*/

	// now draw normal (non-blurred) text over drop-shadow
	cairo_set_source_rgba (cr, 0.952941176, 0.952941176, 0.905882353, 0.6);
	/*cairo_set_source_rgba (cr,
			       TEXT_BODY_COLOR_R,
			       TEXT_BODY_COLOR_G,
			       TEXT_BODY_COLOR_B,
			       TEXT_BODY_COLOR_A);*/
	pango_cairo_show_layout (cr, layout);

	g_object_unref (layout);

	// create the surface/blur-cache from the normal surface
	if (priv->tile_body)
		tile_destroy (priv->tile_body);
	priv->tile_body = tile_new (normal, BUBBLE_CONTENT_BLUR_RADIUS);

	// clean up
	cairo_destroy (cr);
	cairo_surface_destroy (normal);

	// reset refresh-flag
	priv->message_body_needs_refresh = FALSE;
}

void
_refresh_indicator (Bubble* self)
{
	BubblePrivate*   priv   = GET_PRIVATE (self);
	Defaults*        d      = self->defaults;
	cairo_surface_t* normal = NULL;
	cairo_t*         cr     = NULL;

	// create temp. scratch surface
	normal = cairo_image_surface_create (
			CAIRO_FORMAT_ARGB32,
			EM2PIXELS (defaults_get_bubble_width (d), d) -
			3 * EM2PIXELS (defaults_get_margin_size (d), d) -
			EM2PIXELS (defaults_get_icon_size (d), d)
			+ 2 * BUBBLE_CONTENT_BLUR_RADIUS,
			EM2PIXELS (defaults_get_icon_size (d), d) / 5.0f
			+ 2 * BUBBLE_CONTENT_BLUR_RADIUS);
	if (cairo_surface_status (normal) != CAIRO_STATUS_SUCCESS)
		return;

	// create context for normal surface
	cr = cairo_create (normal);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (normal);
		return;
	}

	// clear normal surface
    	cairo_scale (cr, 1.0f, 1.0f);
	cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
	cairo_paint (cr);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	_draw_value_indicator (
		cr,
		priv->value,
		BUBBLE_CONTENT_BLUR_RADIUS,
		BUBBLE_CONTENT_BLUR_RADIUS,
		EM2PIXELS (defaults_get_bubble_width (d), d) -
		3 * EM2PIXELS (defaults_get_margin_size (d), d) -
		EM2PIXELS (defaults_get_icon_size (d), d),
		EM2PIXELS (defaults_get_gauge_size (d), d),
		EM2PIXELS (defaults_get_gauge_outline_width (d), d));

	// create the surface/blur-cache from the normal surface
	if (priv->tile_indicator)
		tile_destroy (priv->tile_indicator);
	priv->tile_indicator = tile_new (normal, BUBBLE_CONTENT_BLUR_RADIUS/2);

	// clean up
	cairo_destroy (cr);
	cairo_surface_destroy (normal);
}

void
_render_background (Bubble*  self,
                    cairo_t* cr,
		    gdouble  alpha_normal,
		    gdouble  alpha_blur)
{
	Defaults* d = self->defaults;

	tile_paint (self->priv->tile_background,
		    cr,
		    0.0f,
		    0.0f,
		    alpha_normal,
		    alpha_blur);

	// layout-grid and urgency-indication bar
	if (g_getenv ("DEBUG"))
	{
		// for debugging layout and positioning issues
		_draw_layout_grid (cr, self);

		switch (bubble_get_urgency (self))
		{
			// low urgency-bar is painted blue
			case 0:
				cairo_set_source_rgb (cr, 0.25f, 0.5f, 1.0f);
			break;

			// normal urgency-bar is painted green
			case 1:
				cairo_set_source_rgb (cr, 0.0f, 1.0f, 0.0f);
			break;

			// urgent urgency-bar is painted red
			case 2:
				cairo_set_source_rgb (cr, 1.0f, 0.0f, 0.0f);
			break;

			default:
			break;
		}

		draw_curved_rect (cr,
			 			  EM2PIXELS (defaults_get_bubble_shadow_size (d), d) + 2.0f,
						  EM2PIXELS (defaults_get_bubble_shadow_size (d), d) + 2.0f,
						  EM2PIXELS (defaults_get_bubble_width (d), d) - 4.0f,
						  2.0f * EM2PIXELS (defaults_get_bubble_shadow_size (d), d) - 2.0f,
						  EM2PIXELS (defaults_get_bubble_corner_radius (d), d) - 2.0f,
						  FALSE);
		/*draw_round_rect (
			cr,
			1.0f,
			EM2PIXELS (defaults_get_bubble_shadow_size (d), d) + 2.0f,
			EM2PIXELS (defaults_get_bubble_shadow_size (d), d) + 2.0f,
			EM2PIXELS (defaults_get_bubble_corner_radius (d), d) - 2.0f,
			EM2PIXELS (defaults_get_bubble_width (d), d) - 4.0f,
			2.0f * EM2PIXELS (defaults_get_bubble_shadow_size (d), d) - 2.0f);*/
		cairo_fill (cr);

		cairo_set_source_rgb (cr, 0.0f, 0.0f, 0.0f);
		cairo_set_font_size (
			cr,
			EM2PIXELS (defaults_get_text_body_size  (d), d));
		cairo_move_to (
			cr,
			EM2PIXELS (defaults_get_text_body_size  (d), d) +
			EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
			2.0f,
			EM2PIXELS (defaults_get_text_body_size  (d), d) +
			EM2PIXELS (defaults_get_bubble_shadow_size (d), d) +
			2.0f +
			((2.0f * EM2PIXELS (defaults_get_bubble_shadow_size (d), d) - 2.0f) -
			EM2PIXELS (defaults_get_text_body_size  (d), d)) / 2);

		switch (bubble_get_urgency (self))
		{
			case 0:
				cairo_show_text (
					cr,
					"low - report incorrect urgency?");
			break;

			case 1:
				cairo_show_text (
					cr,
					"normal - report incorrect urgency?");
			break;

			case 2:
				cairo_show_text (
					cr,
					"urgent - report incorrect urgency?");
			break;

			default:
			break;
		}
	}
}

void
_render_sd_buttons (Bubble*  self,
                    cairo_t* cr,
					gint     x,
					gint     y,
                    gdouble  alpha_normal,
                    gdouble  alpha_blur)
{
	BubblePrivate* priv = GET_PRIVATE (self);

	tile_paint (priv->tile_sd_buttons, cr, x, y, alpha_normal, alpha_blur);
}

void
_render_icon (Bubble*  self,
	      cairo_t* cr,
	      gint     x,
	      gint     y,
	      gdouble  alpha_normal,
	      gdouble  alpha_blur)
{
	BubblePrivate*   priv = GET_PRIVATE (self);
	cairo_pattern_t* pattern;

	tile_paint (priv->tile_icon,
		    cr,
		    x,
		    y,
		    alpha_normal,
		    alpha_blur);
}

void
_render_secondary_icon (Bubble*  self,
						cairo_t* cr,
						gint     x,
						gint     y,
						gdouble  alpha_normal,
						gdouble  alpha_blur)
{
	tile_paint (GET_PRIVATE (self)->tile_secondary_icon,
				cr,
				x,
				y,
				alpha_normal,
				alpha_blur);
}

void
_render_title (Bubble*  self,
	       cairo_t* cr,
	       gint     x,
	       gint     y,
	       gdouble  alpha_normal,
	       gdouble  alpha_blur)
{
	tile_paint (self->priv->tile_title,
		    cr,
		    x,
		    y,
		    alpha_normal,
		    alpha_blur);
}

void
_render_body (Bubble*  self,
	      cairo_t* cr,
	      gint     x,
	      gint     y,
	      gdouble  alpha_normal,
	      gdouble  alpha_blur)
{
	tile_paint (self->priv->tile_body,
		    cr,
		    x,
		    y,
		    alpha_normal,
		    alpha_blur);
}

void
_render_indicator (Bubble*  self,
		   cairo_t* cr,
		   gint     x,
		   gint     y,
		   gdouble  alpha_normal,
		   gdouble  alpha_blur)
{
	BubblePrivate*   priv = GET_PRIVATE (self);
	cairo_pattern_t* pattern;

	tile_paint (priv->tile_indicator,
		    cr,
		    x,
		    y,
		    alpha_normal,
		    alpha_blur);
}

void
_render_layout (Bubble*  self,
		cairo_t* cr,
		gdouble  alpha_normal,
		gdouble  alpha_blur)
{
	Defaults* d           = self->defaults;
	gint      shadow      = EM2PIXELS (defaults_get_bubble_shadow_size (d), d);
	gint      icon_half   = EM2PIXELS (defaults_get_icon_size (d), d) / 2;
	gint      width_half  = EM2PIXELS (defaults_get_bubble_width (d), d) / 2;
	gint      height_half = EM2PIXELS (defaults_get_bubble_min_height (d), d) / 2;
	gint      gauge_half  = EM2PIXELS (defaults_get_gauge_size (d), d) / 2;
	gint      margin      = EM2PIXELS (defaults_get_margin_size (d), d);

	switch (bubble_get_layout (self))
	{
		case LAYOUT_ICON_ONLY:
			_render_icon (self,
				      cr,
				      shadow + width_half - icon_half - BUBBLE_CONTENT_BLUR_RADIUS,
				      shadow + height_half - icon_half - BUBBLE_CONTENT_BLUR_RADIUS,
				      alpha_normal,
				      alpha_blur);
		break;

		case LAYOUT_ICON_INDICATOR:
			_render_icon (self,
				      cr,
				      shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				      shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				      alpha_normal,
				      alpha_blur);
			_render_indicator (self,
					   cr,
					   shadow + 2 * margin + 2 * icon_half - BUBBLE_CONTENT_BLUR_RADIUS,
					   shadow + margin + icon_half - gauge_half - BUBBLE_CONTENT_BLUR_RADIUS,
					   alpha_normal,
					   alpha_blur);
		break;

		case LAYOUT_ICON_TITLE:
			_render_icon (self,
				      cr,
				      shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				      shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				      alpha_normal,
				      alpha_blur);
			_render_title (self,
				       cr,
				       shadow + 2 * margin + 2 * icon_half - BUBBLE_CONTENT_BLUR_RADIUS,
				       shadow + margin + icon_half - self->priv->title_height / 2 - BUBBLE_CONTENT_BLUR_RADIUS,
				       alpha_normal,
				       alpha_blur);
		break;

		case LAYOUT_ICON_TITLE_BODY:
			_render_icon (self,
				      cr,
				      shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				      shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				      alpha_normal,
				      alpha_blur);
			_render_title (self,
				       cr,
				       shadow + 2 * margin + 2 * icon_half - BUBBLE_CONTENT_BLUR_RADIUS,
				       shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				       alpha_normal,
				       alpha_blur);
			_render_body (self,
				      cr,
				      shadow + 2 * margin + 2 * icon_half - BUBBLE_CONTENT_BLUR_RADIUS,
				      shadow + margin + GET_PRIVATE (self)->title_height - BUBBLE_CONTENT_BLUR_RADIUS,
				      alpha_normal,
				      alpha_blur);
		break;

		case LAYOUT_TITLE_BODY:
			_render_title (self,
				       cr,
				       shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				       shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				       alpha_normal,
				       alpha_blur);
			_render_body (self,
				      cr,
				      shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				      shadow + margin + GET_PRIVATE (self)->title_height - BUBBLE_CONTENT_BLUR_RADIUS,
				      alpha_normal,
				      alpha_blur);
		break;

		case LAYOUT_TITLE_ONLY:
			_render_title (self,
				       cr,
				       shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				       shadow + margin - BUBBLE_CONTENT_BLUR_RADIUS,
				       alpha_normal,
				       alpha_blur);
		break;

		case LAYOUT_NONE:
			// should be intercepted by stack_notify_handler()
			g_warning ("WARNING: No layout defined!!!\n");
		break;
	}
}

// the behind-bubble blur only works with the enabled/working compiz-plugin blur
// by setting the hint _COMPIZ_WM_WINDOW_BLUR on the bubble-window
void
_set_bg_blur (GtkWidget* window,
	      gboolean   set_blur,
	      gint       shadow_size)
{}

static
void
screen_changed_handler (GtkWindow* window,
						GdkScreen* old_screen,
						gpointer   data)
{}

static
void
update_input_shape (GtkWidget* window)
{}

static void
update_shape (Bubble* self)
{}

static void
composited_changed_handler (GtkWidget* window,
			    gpointer   data)
{
	Bubble* bubble;

	bubble = (Bubble*) G_OBJECT (data);

	GET_PRIVATE (bubble)->composited = TRUE;
}

static
gboolean
expose_handler (GtkWidget*      window,
		GdkEventExpose* event,
		gpointer        data)
{
	return TRUE;
}

static gboolean
redraw_handler (Bubble* bubble)
{
	return TRUE;
}

static
GdkPixbuf*
load_icon (const gchar* filename,
	   gint         icon_size)
{
	GdkPixbuf*    buffer = NULL;
	GdkPixbuf*    pixbuf = NULL;
	GtkIconTheme* theme  = NULL;
	GError*       error  = NULL;

	/* sanity check */
	g_return_val_if_fail (filename, NULL);

	/* Images referenced must always be local files. */
	if (!strncmp (filename, "file://", 7))
		filename += 7;

	if (filename[0] == '/')
	{
		/* load image into pixbuf */
		pixbuf = gdk_pixbuf_new_from_file_at_scale (filename,
							    icon_size,
							    icon_size,
							    TRUE,
							    NULL);
	} else {
		/* TODO: rewrite, check for SVG support, raise apport
		** notification for low-res icons */
		theme = defaults_get_icon_theme(defaults_get_default());
		buffer = gtk_icon_theme_load_icon (theme,
						   filename,
                                                   icon_size,
						   (GtkIconLookupFlags)(GTK_ICON_LOOKUP_FORCE_SVG |
						   GTK_ICON_LOOKUP_FORCE_SVG |
						   GTK_ICON_LOOKUP_GENERIC_FALLBACK |
					   GTK_ICON_LOOKUP_FORCE_SIZE),
						   &error);
		if (error)
		{
			g_print ("loading icon '%s' caused error: '%s'",
				 filename,
				 error->message);
			g_error_free (error);
			error = NULL;
			pixbuf = NULL;
		}
		else
		{
			/* copy and unref buffer so on an icon-theme change old
			** icons are not kept in memory due to dangling
			** references, self also makes sure we do not need to
			** connect to GtkWidget::style-set signal for the
			** GdkPixbuf we get from gtk_icon_theme_load_icon() */
			pixbuf = gdk_pixbuf_copy (buffer);
			g_object_unref (buffer);
		}
	}

	return pixbuf;
}

static
gboolean
pointer_update (Bubble* bubble)
{
	return TRUE;
}

//-- internal API --------------------------------------------------------------

static void
bubble_dispose (GObject* gobject)
{
	BubblePrivate* priv;

	if (!gobject || !IS_BUBBLE (gobject))
		return;

	priv = GET_PRIVATE (gobject);

	if (GTK_IS_WIDGET (priv->widget))
	{
		gtk_widget_destroy (GTK_WIDGET (priv->widget));
		priv->widget = NULL;
	}

    	if (priv->title)
	{
		g_string_free (priv->title,
			       TRUE);
		priv->title = NULL;
	}

    	if (priv->message_body)
	{
		g_string_free (priv->message_body,
			       TRUE);
		priv->message_body = NULL;
	}

    	if (priv->synchronous)
	{
		g_free ((gpointer) priv->synchronous);
		priv->synchronous = NULL;
	}

	if (priv->icon_pixbuf)
	{
		g_object_unref (priv->icon_pixbuf);
		priv->icon_pixbuf = NULL;
	}

	if (priv->sd_answer_pixbuf)
	{
		g_object_unref (priv->sd_answer_pixbuf);
		priv->sd_answer_pixbuf = NULL;
	}

	if (priv->sd_ignore_pixbuf)
	{
		g_object_unref (priv->sd_ignore_pixbuf);
		priv->sd_ignore_pixbuf = NULL;
	}

	if (priv->pointer_update_id)
	{
		g_source_remove (priv->pointer_update_id);
		priv->pointer_update_id = 0;
	}

	if (priv->draw_handler_id)
	{
		g_source_remove (priv->draw_handler_id);
		priv->draw_handler_id = 0;
	}

	if (priv->timer_id)
	{
		g_source_remove (priv->timer_id);
		priv->timer_id = 0;
	}

	if (priv->tile_background_part)
	{
		tile_destroy (priv->tile_background_part);
		priv->tile_background_part = NULL;
	}

	if (priv->tile_background)
	{
		tile_destroy (priv->tile_background);
		priv->tile_background = NULL;
	}

	if (priv->tile_icon)
	{
		tile_destroy (priv->tile_icon);
		priv->tile_icon = NULL;
	}

	if (priv->tile_secondary_icon)
	{
		tile_destroy (priv->tile_secondary_icon);
		priv->tile_secondary_icon = NULL;
	}

	if (priv->tile_title)
	{
		tile_destroy (priv->tile_title);
		priv->tile_title = NULL;
	}

	if (priv->tile_body)
	{
		tile_destroy (priv->tile_body);
		priv->tile_body = NULL;
	}

	if (priv->tile_indicator)
	{
		tile_destroy (priv->tile_indicator);
		priv->tile_indicator = NULL;
	}

	if (priv->old_icon_filename)
	{
		g_string_free (priv->old_icon_filename, TRUE);
		priv->old_icon_filename = NULL;
	}

	if (priv->scratch)
	{
		cairo_surface_destroy (priv->scratch);
	}

	if (priv->secondary_icon_pixbuf)
	{
		g_object_unref (priv->secondary_icon_pixbuf);
		priv->secondary_icon_pixbuf = NULL;
	}

	if (priv->sender)
	{
		g_string_free (priv->sender, TRUE);
		priv->sender = NULL;
	}

	// chain up to the parent class
	G_OBJECT_CLASS (bubble_parent_class)->dispose (gobject);
}

static void
bubble_finalize (GObject* gobject)
{
	// chain up to the parent class
	G_OBJECT_CLASS (bubble_parent_class)->finalize (gobject);
}

static void
bubble_init (Bubble* self)
{
	// If you need specific construction properties to complete
	// initialization, delay initialization completion until the
	// property is set.

	BubblePrivate *priv;

	self->priv = priv                = GET_PRIVATE (self);
	priv->layout                     = LAYOUT_NONE;
	priv->title                      = NULL;
	priv->message_body               = NULL;
	priv->title_needs_refresh        = FALSE;
	priv->message_body_needs_refresh = FALSE;
	priv->visible                    = FALSE;
	priv->icon_pixbuf                = NULL;
	priv->value                      = -2;
	priv->synchronous                = NULL;
	priv->sender                     = g_string_new("");
	priv->draw_handler_id            = 0;
	priv->pointer_update_id          = 0;
}

static void
bubble_get_property (GObject*    gobject,
		     guint       prop,
		     GValue*     value,
		     GParamSpec* spec)
{
	switch (prop)
	{
		default :
			G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop, spec);
		break;
	}
}

static void
bubble_class_init (BubbleClass* klass)
{
	GObjectClass* gobject_class = G_OBJECT_CLASS (klass);

	g_type_class_add_private (klass, sizeof (BubblePrivate));

	gobject_class->dispose      = bubble_dispose;
	gobject_class->finalize     = bubble_finalize;
	gobject_class->get_property = bubble_get_property;

	g_bubble_signals[TIMED_OUT] = g_signal_new (
		"timed-out",
		G_OBJECT_CLASS_TYPE (gobject_class),
		G_SIGNAL_RUN_LAST,
		G_STRUCT_OFFSET (BubbleClass, timed_out),
		NULL,
		NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE,
		0);

	g_bubble_signals[VALUE_CHANGED] = g_signal_new (
		"value-changed",
		G_OBJECT_CLASS_TYPE (gobject_class),
		G_SIGNAL_RUN_LAST,
		G_STRUCT_OFFSET (BubbleClass, value_changed),
		NULL,
		NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE,
		1,
        G_TYPE_INT);

    g_bubble_signals[MESSAGE_BODY_DELETED] = g_signal_new (
		"message-body-deleted",
		G_OBJECT_CLASS_TYPE (gobject_class),
		G_SIGNAL_RUN_LAST,
		G_STRUCT_OFFSET (BubbleClass, message_body_deleted),
		NULL,
		NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE,
		1,
        G_TYPE_STRING);

    g_bubble_signals[MESSAGE_BODY_INSERTED] = g_signal_new (
		"message-body-inserted",
		G_OBJECT_CLASS_TYPE (gobject_class),
		G_SIGNAL_RUN_LAST,
		G_STRUCT_OFFSET (BubbleClass, message_body_inserted),
		NULL,
		NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE,
		1,
        G_TYPE_STRING);
}

//-- public API ----------------------------------------------------------------

Bubble*
bubble_new (Defaults* defaults)
{
	Bubble*        self        = NULL;
	GtkWidget*     window      = NULL;
	BubblePrivate* priv;
	GdkRGBA        transparent = {0.0, 0.0, 0.0, 0.0};

	self = (Bubble*)g_object_new (BUBBLE_TYPE, NULL);
	if (!self)
		return NULL;

	self->defaults = defaults;
	priv = GET_PRIVATE (self);

	//g_object_set_data (G_OBJECT(window), "bubble", (gpointer) self);

	// TODO: fold some of that back into bubble_init
	self->priv = GET_PRIVATE (self);
	self->priv->layout                     = LAYOUT_NONE;
	self->priv->widget                     = window;
	self->priv->title                      = g_string_new ("");
	self->priv->message_body               = g_string_new ("");
	self->priv->title_needs_refresh        = FALSE;
	self->priv->message_body_needs_refresh = FALSE;
	self->priv->icon_pixbuf                = NULL;
	self->priv->sd_ignore_pixbuf           = NULL;
	self->priv->sd_answer_pixbuf           = NULL;
	self->priv->value                      = -2;
	self->priv->visible                    = FALSE;
	self->priv->timeout                    = 5000;
	self->priv->mouse_over                 = FALSE;
	self->priv->distance                   = 1.0f;
	self->priv->title_width                = 0;
	self->priv->title_height               = 0;
	self->priv->body_width                 = 0;
	self->priv->body_height                = 0;
	self->priv->append                     = FALSE;
	self->priv->icon_only                  = FALSE;
	self->priv->tile_background_part       = NULL;
	self->priv->tile_background            = NULL;
	self->priv->tile_icon                  = NULL;
	self->priv->tile_secondary_icon        = NULL;
	self->priv->tile_title                 = NULL;
	self->priv->tile_body                  = NULL;
	self->priv->tile_indicator             = NULL;
	self->priv->prevent_fade               = FALSE;
	self->priv->old_icon_filename          = g_string_new ("");
	self->priv->scratch                    = NULL;
	self->priv->secondary_icon_pixbuf      = NULL;
	self->priv->_units                     = Units::getInstance ();

	return self;
}

void
bubble_set_notification (Bubble*	   self,
						 Notification* n)
{
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self))
		return;

	priv = GET_PRIVATE (self);

	priv->n			   = n;
	priv->icon_pixbuf  = (GdkPixbuf*) g_object_ref (notification_get_icon_pixbuf (priv->n));
	priv->value        = notification_get_value (priv->n);
	priv->icon_only    = notification_get_icon_only_hint (priv->n);
	priv->id           = notification_get_id (priv->n);
    priv->secondary_icon_pixbuf = (GdkPixbuf*) g_object_ref (notification_get_secondary_icon (priv->n));

    if (notification_get_num_actions(priv->n) >= 4)
    {
    	priv->sd_answer_pixbuf = gdk_pixbuf_new_from_file ("/usr/share/notify-osd/answer_button.png", NULL);
    	priv->sd_ignore_pixbuf = gdk_pixbuf_new_from_file ("/usr/share/notify-osd/ignore_button.png", NULL);
    }
	g_string_assign (priv->title, notification_get_summary (priv->n));
	g_string_assign (priv->message_body, notification_get_body (priv->n));
	g_string_assign (priv->sender, notification_get_dbus_sender (priv->n));
}

gchar*
bubble_get_synchronous (Bubble* self)
{
	g_return_val_if_fail (IS_BUBBLE (self), NULL);

	return GET_PRIVATE (self)->synchronous;
}

gchar*
bubble_get_sender (Bubble* self)
{
	g_return_val_if_fail (IS_BUBBLE (self), NULL);

	return GET_PRIVATE (self)->sender->str;
}

void
bubble_set_title (Bubble*      self,
		  const gchar* title)
{
	gchar*         text;
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self))
		return;

	priv = GET_PRIVATE (self);

	// convert any newline to space
	text = newline_to_space (title);

	if (priv->title)
	{
		if (g_strcmp0 (priv->title->str, text))
			priv->title_needs_refresh = TRUE;

		g_string_free (priv->title, TRUE);
	}

	priv->title = g_string_new (text);

	g_free (text);
}

const gchar*
bubble_get_title (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return NULL;

	return GET_PRIVATE (self)->title->str;
}

void
bubble_set_message_body (Bubble*      self,
			 const gchar* body)
{
	gchar*         text;
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self))
		return;

	priv = GET_PRIVATE (self);

	// filter out any HTML/markup if possible
	text = filter_text (body);

	if (priv->message_body)
		if (g_strcmp0 (priv->message_body->str, text))
			priv->message_body_needs_refresh = TRUE;

	if (priv->message_body && priv->message_body->len != 0)
	{
		g_signal_emit (self,
			       g_bubble_signals[MESSAGE_BODY_DELETED], 
			       0,
			       priv->message_body->str);
		g_string_free (priv->message_body, TRUE);
	}

	priv->message_body = g_string_new (text);

	g_signal_emit (self, g_bubble_signals[MESSAGE_BODY_INSERTED], 0, text);

	g_free (text);
}

const gchar*
bubble_get_message_body (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return NULL;

	return GET_PRIVATE (self)->message_body->str;
}

void
bubble_set_icon_from_path (Bubble*      self,
			   const gchar* filepath)
{
	Defaults*      d;
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self) || !g_strcmp0 (filepath, ""))
		return;

	priv = GET_PRIVATE (self);

	// check if an app tries to set the same file as icon again, self check
	// avoids superfluous regeneration of the tile/blur-cache for the icon,
	// thus it improves performance in update- and append-cases
	if (!g_strcmp0 (priv->old_icon_filename->str, filepath))
		return;

	// store the new icon-basename
	g_string_assign (priv->old_icon_filename, filepath);

	if (priv->icon_pixbuf)
	{
		g_object_unref (priv->icon_pixbuf);
		priv->icon_pixbuf = NULL;
	}

	d = self->defaults;
	priv->icon_pixbuf = load_icon (filepath,
				       EM2PIXELS (defaults_get_icon_size (d), d));

	_refresh_icon (self);
}

void
bubble_set_icon (Bubble*      self,
		 const gchar* filename)
{
	Defaults*      d;
	BubblePrivate* priv;
#ifdef TEMPORARY_ICON_PREFIX_WORKAROUND
	gchar*         notify_osd_iconname;
#endif

 	if (!self || !IS_BUBBLE (self) || !g_strcmp0 (filename, ""))
		return;

	priv = GET_PRIVATE (self);

	//basename = g_path_get_basename (filename);

	// check if an app tries to set the same file as icon again, self check
	// avoids superfluous regeneration of the tile/blur-cache for the icon,
	// thus it improves performance in update- and append-cases
	if (!g_strcmp0 (priv->old_icon_filename->str, filename))
		return;

	// store the new icon-basename
	g_string_assign (priv->old_icon_filename, filename);

	if (priv->icon_pixbuf)
	{
		g_object_unref (priv->icon_pixbuf);
		priv->icon_pixbuf = NULL;
	}

	d = self->defaults;

#ifdef TEMPORARY_ICON_PREFIX_WORKAROUND
	notify_osd_iconname = g_strdup_printf (NOTIFY_OSD_ICON_PREFIX "-%s",
					       filename);
	priv->icon_pixbuf = load_icon (notify_osd_iconname,
				       EM2PIXELS (defaults_get_icon_size (d),
						  d));
	g_free (notify_osd_iconname);
#endif

	// fallback to non-notify-osd name
	if (!priv->icon_pixbuf)
		priv->icon_pixbuf = load_icon (filename,
					       EM2PIXELS (defaults_get_icon_size (d), d));

	_refresh_icon (self);
}

static GdkPixbuf *
scale_pixbuf (const GdkPixbuf *pixbuf, gint size)
{
	GdkPixbuf *scaled_icon;
	GdkPixbuf *new_icon;
	gint w, h, dest_x, dest_y, new_width, new_height, max_edge;

	dest_x = dest_y = 0;

	w = gdk_pixbuf_get_width (pixbuf);
	h = gdk_pixbuf_get_height (pixbuf);

	max_edge = MAX (w, h);

	// temporarily cast to float so we don't end up missing fractional parts
	// from the division, especially nasty for 0.something :)
	// e.g.: 99 / 100 = 0 but 99.0 / 100.0 = 0.99
	new_width = size * ((gfloat) w / (gfloat) max_edge);
	new_height =  size * ((gfloat) h / (gfloat) max_edge);

	/* Scale the pixbuf down, preserving the aspect ratio */
	scaled_icon = gdk_pixbuf_scale_simple (pixbuf,
					       new_width,
					       new_height,
					       GDK_INTERP_BILINEAR);

	if (w == h)
		return scaled_icon;

	/* Create a square pixbuf with an alpha channel */
	new_icon = gdk_pixbuf_new (gdk_pixbuf_get_colorspace (scaled_icon),
				   TRUE,
				   gdk_pixbuf_get_bits_per_sample (scaled_icon),
				   size, size);

	/* Clear the pixbuf so it is transparent */
	gdk_pixbuf_fill (new_icon, 0x00000000);

	/* Center the rectangular pixbuf inside the transparent square */
	if (new_width > new_height)
		dest_y = (new_width - new_height) / 2;
	else
		dest_x = (new_height - new_width) / 2;

	/* Copy the rectangular pixbuf into the new pixbuf at a centered position */
	gdk_pixbuf_copy_area (scaled_icon,
			      0, 0,
			      gdk_pixbuf_get_width (scaled_icon),
			      gdk_pixbuf_get_height (scaled_icon),
			      new_icon,
			      dest_x, dest_y);
	g_object_unref (scaled_icon);

	return new_icon;
}

void
bubble_set_icon_from_pixbuf (Bubble*    self,
			     GdkPixbuf* pixbuf)
{
	GdkPixbuf*     scaled;
	gint           height;
	gint           width;
	Defaults*      d;
	BubblePrivate* priv;

 	if (!self || !IS_BUBBLE (self) || !pixbuf)
		return;

	priv = GET_PRIVATE (self);

	// "reset" the stored the icon-filename, fixes LP: #451086
	g_string_assign (priv->old_icon_filename, "\0");

	if (priv->icon_pixbuf)
	{
		g_object_unref (priv->icon_pixbuf);
		priv->icon_pixbuf = NULL;
	}

	height = gdk_pixbuf_get_height (pixbuf);
	width = gdk_pixbuf_get_width (pixbuf);

	d = self->defaults;

	if (width != defaults_get_icon_size (d) ||
            height != defaults_get_icon_size (d))
	{
		scaled = scale_pixbuf (pixbuf, 6 * GET_PRIVATE (self)->_units->getGU ());
		g_object_unref (pixbuf);
		pixbuf = scaled;
	}

	priv->icon_pixbuf = pixbuf;

	_refresh_icon (self);
}

GdkPixbuf*
bubble_get_icon_pixbuf (Bubble *self)
{
	g_return_val_if_fail (IS_BUBBLE (self), NULL);

	return GET_PRIVATE (self)->icon_pixbuf;
}

void
bubble_set_value (Bubble* self,
		  gint    value)
{
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self))
		return;

	priv = GET_PRIVATE (self);

	// only really cause a refresh (blur, recreating tile-/blur-cache) if
	// a different value has been set, self helps improve performance when
	// a user e.g. presses and holds down keys for Volume-Up/Down or
	// Brightness-Up/Down and apps like gnome-settings-daemon or
	// gnome-power-daemon fire continues notification-updates due to the
	// key-repeat events
	if (priv->value != value)
	{
		priv->value = value;
		g_signal_emit (self, g_bubble_signals[VALUE_CHANGED], 0, value);
		_refresh_indicator (self);
	}
}

gint
bubble_get_value (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return -2;

	return GET_PRIVATE (self)->value;
}

void
bubble_set_size (Bubble* self,
		 gint    width,
		 gint    height)
{
	if (!self || !IS_BUBBLE (self))
		return;

	GET_PRIVATE (self)->width = width;
	GET_PRIVATE (self)->height = height;
}

void
bubble_get_size (Bubble* self,
		 gint*   width,
		 gint*   height)
{
	if (!self || !IS_BUBBLE (self))
		return;

	*width = GET_PRIVATE (self)->width;
	*height = GET_PRIVATE (self)->height;
}

void
bubble_set_timeout (Bubble* self,
		    guint   timeout)
{
	if (!self || !IS_BUBBLE (self))
		return;

	GET_PRIVATE (self)->timeout = timeout;
}

/* a timeout of 0 doesn't make much sense now does it, thus 0 indicates an
** error */
guint
bubble_get_timeout (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return 0;

	return GET_PRIVATE(self)->timeout;
}

void
bubble_set_timer_id (Bubble* self,
		     guint   timer_id)
{
	if (!self || !IS_BUBBLE (self))
		return;

	GET_PRIVATE(self)->timer_id = timer_id;
}

/* a valid GLib timer-id is always > 0, thus 0 indicates an error */
guint
bubble_get_timer_id (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return 0;

	return GET_PRIVATE(self)->timer_id;
}

void
bubble_set_mouse_over (Bubble*  self,
		       gboolean flag)
{
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self))
		return;

	priv = GET_PRIVATE (self);

	/* did anything change? */
	if (priv->mouse_over != flag)
	{
		priv->mouse_over = flag;
		//update_shape (self);

		if (flag) {
			g_debug("mouse entered bubble, supressing timeout");
			bubble_clear_timer(self);
		} else {
			g_debug("mouse left bubble, continuing timeout");
			bubble_set_timeout(self, 3000);
			bubble_start_timer(self, TRUE);
		}
        
	}

}

gboolean
bubble_is_mouse_over (Bubble* self)
{
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self))
		return FALSE;

	priv = GET_PRIVATE (self);

	if (priv->prevent_fade)
		return FALSE;

	return priv->mouse_over;
}

void
bubble_move (Bubble* self,
	     gint    x,
	     gint    y)
{
	if (!self || !IS_BUBBLE (self))
		return;
}

void
bubble_show (Bubble* self)
{}

/* mostly called when we change the content of the bubble
   and need to force a redraw
*/
void
bubble_refresh (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return;
}

static inline gboolean
bubble_is_composited (Bubble *bubble)
{
	g_return_val_if_fail(bubble && IS_BUBBLE(bubble), false);
	return GET_PRIVATE (bubble)->composited;
}

static inline GtkWindow*
bubble_get_window (Bubble *bubble)
{
	return NULL;
}

gboolean
bubble_timed_out (Bubble* self)
{
	g_return_val_if_fail (IS_BUBBLE (self), FALSE);

	dbus_send_close_signal (bubble_get_sender (self),
				bubble_get_id (self),
				1);
	g_signal_emit (self, g_bubble_signals[TIMED_OUT], 0);

	g_object_unref (self);

	return FALSE;
}

void
bubble_set_id (Bubble* self,
	       guint   id)
{
	if (!self || !IS_BUBBLE (self))
		return;

	GET_PRIVATE (self)->id = id;
}

guint
bubble_get_id (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return 0;

	return GET_PRIVATE (self)->id;
}

gboolean
bubble_is_visible (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return FALSE;

	return GET_PRIVATE (self)->visible;
}

void
bubble_start_timer (Bubble*  self,
		    gboolean trigger)
{
	guint          timer_id;
	BubblePrivate* priv;

	if (!self || !IS_BUBBLE (self))
		return;

	priv = GET_PRIVATE (self);

	timer_id = bubble_get_timer_id (self);
	if (timer_id > 0)
		g_source_remove (timer_id);

	/* and now let the timer tick... we use g_timeout_add_seconds() here
	** because for the bubble timeouts microsecond-precise resolution is not
	** needed, furthermore self also allows glib more room for optimizations
	** and improve system-power-usage to be more efficient */
	bubble_set_timer_id (
		self,
		g_timeout_add_seconds (bubble_get_timeout (self) / 1000,
				       (GSourceFunc) bubble_timed_out,
				       self));

	/* if the bubble is displaying a value that is out of bounds
	   trigger a dim/glow animation */
	/*if (trigger)
		if (priv->value == -1 || priv->value == 101)
			bubble_start_glow_effect (self, 500);*/
}

void
bubble_clear_timer (Bubble* self) 
{
	guint timer_id;

	if (!self || !IS_BUBBLE (self))
		return;
    
	timer_id = GET_PRIVATE(self)->timer_id;

	if (timer_id > 0) {
		g_source_remove (timer_id);
		bubble_set_timer_id(self, 0);
	}
}

void
bubble_get_position (Bubble* self,
		     gint*   x,
		     gint*   y)
{
  *x = NULL;
  *y = NULL;

	if (!self || !IS_BUBBLE (self))
		return;
}

gint
bubble_get_height (Bubble *self)
{
 return 100;
}

gint
bubble_get_future_height (Bubble *self)
{
	if (!self || !IS_BUBBLE (self))
		return 0;

	return GET_PRIVATE (self)->future_height;
}

gint
_calc_title_height (Bubble* self,
		    gint    title_width /* requested text-width in pixels */)
{
	Defaults*             d;
	cairo_surface_t*      surface;
	cairo_t*              cr;
	PangoFontDescription* desc    = NULL;
	PangoLayout*          layout  = NULL;
	PangoRectangle        log_rect = {0, 0, 0, 0};
	gint                  title_height;
	BubblePrivate*        priv;
	gchar*                text_font_face = NULL;

	if (!self || !IS_BUBBLE (self))
		return 0;

	d    = self->defaults;
	priv = GET_PRIVATE (self);

	surface = cairo_image_surface_create (CAIRO_FORMAT_A1, 1, 1);
	if (cairo_surface_status (surface) != CAIRO_STATUS_SUCCESS) {
		if (surface)
			cairo_surface_destroy (surface);
		return 0;
	}

	cr = cairo_create (surface);
	cairo_surface_destroy (surface);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS) {
		if (cr)
			cairo_destroy (cr);
		return 0;
	}

	layout = pango_cairo_create_layout (cr);
	//text_font_face = defaults_get_text_font_face (d);
	desc = pango_font_description_from_string ("Ubuntu Medium");
	//g_free ((gpointer) text_font_face);

	// make sure system-wide font-options like hinting, antialiasing etc.
	// are taken into account
	pango_cairo_context_set_resolution  (pango_layout_get_context (layout),
					     defaults_get_screen_dpi (d));
	pango_layout_context_changed (layout);

	pango_font_description_set_absolute_size (desc,
											  priv->_units->sizeToPixel ("medium") * PANGO_SCALE);
	/*pango_font_description_set_size (desc,
					 defaults_get_system_font_size (d) *
					 defaults_get_text_title_size (d) *
					 PANGO_SCALE);*/

	/*pango_font_description_set_weight (
		desc,
		(PangoWeight)defaults_get_text_title_weight (d));*/

	pango_layout_set_font_description (layout, desc);
	pango_font_description_free (desc);

	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
	pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
	pango_layout_set_width (layout, title_width * PANGO_SCALE);
	pango_layout_set_height (layout, -1);

	pango_layout_set_text (layout, priv->title->str, priv->title->len);

	pango_layout_get_extents (layout, NULL, &log_rect);
	title_height = PANGO_PIXELS (log_rect.height);
	g_object_unref (layout);
	cairo_destroy (cr);

	return title_height;
}

gint
_calc_body_height (Bubble* self,
		   gint    body_width /* requested text-width in pixels */)
{
	Defaults*             d;
	cairo_t*              cr      = NULL;
	cairo_surface_t*      surf    = NULL;
	PangoFontDescription* desc    = NULL;
	PangoLayout*          layout  = NULL;
	PangoRectangle        log_rect;
	gint                  body_height;
	BubblePrivate*        priv;
	gchar*                text_font_face = NULL;

	if (!self || !IS_BUBBLE (self))
		return 0;

	d    = self->defaults;
	priv = GET_PRIVATE (self);

	//cr = gdk_cairo_create (gtk_widget_get_window (priv->widget));

	// create dummy cairo-surface/context for calculating the text width/height
    surf = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 1, 1);
    cr = cairo_create (surf);

	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS) {
		if (cr)
			cairo_destroy (cr);
		return 0;
	}

	layout = pango_cairo_create_layout (cr);
	//text_font_face = defaults_get_text_font_face (d);
	desc = pango_font_description_from_string ("Ubuntu");
	//g_free ((gpointer) text_font_face);

	// make sure system-wide font-options like hinting, antialiasing etc.
	// are taken into account
	pango_cairo_context_set_resolution  (pango_layout_get_context (layout),
					     defaults_get_screen_dpi (d));
	pango_layout_context_changed (layout);

	pango_font_description_set_absolute_size (desc,
											  priv->_units->sizeToPixel ("small") * PANGO_SCALE);
	/*pango_font_description_set_size (desc,
					 defaults_get_system_font_size (d) *
					 defaults_get_text_body_size (d) *
					 PANGO_SCALE);*/

	/*pango_font_description_set_weight (
		desc,
		(PangoWeight)defaults_get_text_body_weight (d));*/

	pango_layout_set_font_description (layout, desc);

	pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
	pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_MIDDLE);
	pango_layout_set_width (layout, body_width * PANGO_SCALE);
	pango_layout_set_height (layout, -10);

	pango_layout_set_text (layout,
			       priv->message_body->str,
			       priv->message_body->len);

	/* enforce the 10 line-limit, usually only triggered by appended
	** body-message text due to the newline-characters added with each
	** append */
	if (pango_layout_get_line_count (layout) > 10)
	{
		/* get it down to 9 lines, because we add our own ...-line */
		while (pango_layout_get_line_count (layout) > 9)
		{
			GString* string = NULL;

			/* cut leading chunk of text up to the first newline */
			string = g_string_new (g_strstr_len (priv->message_body->str,
							     priv->message_body->len,
							     "\n"));

			/* also cut the first newline itself */
			string = g_string_erase (string, 0, 1);

			/* copy that stripped text back to the body-message */
			g_string_assign (priv->message_body, string->str);

			/* set the new body-message text to the pango-layout */
			pango_layout_set_text (layout,
					       priv->message_body->str,
					       priv->message_body->len);

			/* clean up */
			g_string_free (string, TRUE);
		}

		/* add our own ellipsize-line */
		g_string_prepend (priv->message_body, "...\n");

		/* set final body-message text to the pango-layout again */
		pango_layout_set_text (layout,
				       priv->message_body->str,
				       priv->message_body->len);
	}

	pango_layout_get_extents (layout, NULL, &log_rect);
	body_height = PANGO_PIXELS (log_rect.height);

	pango_font_description_free (desc);
	g_object_unref (layout);
	cairo_destroy (cr);
	cairo_surface_destroy (surf);

	return body_height;
}

void
bubble_recalc_size (Bubble *self)
{
	gint           old_bubble_width  = 0;
	gint           old_bubble_height = 0;
	gint           new_bubble_width  = 0;
	gint           new_bubble_height = 0;
	Defaults*      d;
	BubblePrivate* priv;
	gint           gu                = GET_PRIVATE (self)->_units->getGU ();
	gdouble        icon_height       = 6 * gu;

	if (!self || !IS_BUBBLE (self))
		return;

	d    = self->defaults;
	priv = GET_PRIVATE (self);

	/* FIXME: a quick fix to rescale an icon (e.g. user changed font-size or
	** DPI while a bubble is displayed, thus bubble is re-rendered and the
	** icon needs to adapt to the new size) */
	if (priv->icon_pixbuf)
	{
		GdkPixbuf *pixbuf = NULL;
		if (bubble_get_layout (self) == LAYOUT_ICON_TITLE)
		{
			pixbuf = gdk_pixbuf_scale_simple (priv->icon_pixbuf,
											  2 * gu,
											  2 * gu,
											  GDK_INTERP_BILINEAR);			
		}
		else
		{
			pixbuf = gdk_pixbuf_scale_simple (priv->icon_pixbuf,
											  6 * gu,
											  6 * gu,
											  GDK_INTERP_BILINEAR);			
		}
		g_object_unref (priv->icon_pixbuf);
		priv->icon_pixbuf = pixbuf;
	}

	bubble_determine_layout (self);

	new_bubble_width = 38 * gu;

	switch (priv->layout)
	{
		case LAYOUT_ICON_ONLY:
		case LAYOUT_ICON_INDICATOR:
		case LAYOUT_ICON_TITLE:
			priv->title_width = 33 * gu;

			priv->title_height = _calc_title_height (
					self,
					GET_PRIVATE (self)->title_width);

			new_bubble_height = 6 * gu;
		break;

		case LAYOUT_ICON_TITLE_BODY:
		{
			gdouble available_height = 0.0f;
			gdouble bubble_height    = 0.0f;

			priv->title_width = 26 * gu;
			priv->title_height = _calc_title_height (
					self,
					GET_PRIVATE (self)->title_width);

			priv->body_width = priv->title_width;

			priv->body_height = _calc_body_height (
					self,
					priv->body_width);

			if (icon_height > priv->body_height + priv->title_height + 2 * gu)
				new_bubble_height = icon_height + 4 * gu;
			else
				new_bubble_height = priv->body_height + priv->title_height + 5 * gu;

			if (notification_get_num_actions(priv->n) >= 4)
				new_bubble_height += 6 * gu;
		}
		break;

		case LAYOUT_TITLE_BODY:
		{
			gdouble available_height = 0.0f;
			gdouble bubble_height    = 0.0f;

			priv->title_width = 26 * gu;
			priv->title_height = _calc_title_height (
					self,
					priv->title_width);

			priv->body_width = priv->title_width;
			priv->body_height = _calc_body_height (
					self,
					priv->body_width);

			new_bubble_height = priv->body_height + priv->title_height + 5 * gu;
		}
		break;

		case LAYOUT_TITLE_ONLY:
		{
			priv->title_width = 26 * gu;
			priv->title_height = _calc_title_height (
					self,
					priv->title_width);

			new_bubble_height = priv->title_height + 2 * gu + 2 * gu;
		}
		break;

		case LAYOUT_NONE:
			g_warning ("bubble_recalc_size(): WARNING, no layout!!!\n");
		break;
	}
	priv->future_height = new_bubble_height;

	bubble_get_size (self, &old_bubble_width, &old_bubble_height);
	bubble_set_size (self, new_bubble_width, new_bubble_height);

	// don't refresh tile/blur-caches for background, title or body if the
	// size or contents have not changed, self improves performance for the
	// update- and replace-cases
	/*if (old_bubble_width != new_bubble_width ||
	    old_bubble_height != new_bubble_height)
		_refresh_background (self);

	if (priv->title_needs_refresh)
		_refresh_title (self);

	if (priv->message_body_needs_refresh)
		_refresh_body (self);*/

	_refresh_background (self);
	_refresh_icon (self);
	_refresh_secondary_icon (self);
	_refresh_title (self);
	_refresh_body (self);
	_refresh_sd_buttons (self);

	if (priv->scratch)
	{
		cairo_surface_destroy (priv->scratch);
		priv->scratch = NULL;
	}
	priv->scratch = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
												new_bubble_width,
												new_bubble_height);
	cairo_t* cr = cairo_create (priv->scratch);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	tile_paint (priv->tile_background, cr, 0, 0, 1.0, 0.0);
	tile_paint (priv->tile_icon, cr, 1 * gu, 2 * gu, 1.0, 0.0);

	if (bubble_get_layout (self) == LAYOUT_ICON_TITLE)
		tile_paint (priv->tile_title, cr, 4 * gu, 2 * gu, 1.0, 0.0);
	else
		tile_paint (priv->tile_title, cr, 11 * gu, 2 * gu, 1.0, 0.0);

	tile_paint (priv->tile_body, cr, 11 * gu, priv->title_height + 3 * gu, 1.0, 0.0);
	tile_paint (priv->tile_secondary_icon, cr, 8 * gu, 2 * gu, 1.0, 0.0);
	if (notification_get_num_actions(priv->n) >= 4)
		tile_paint (priv->tile_sd_buttons, cr, 1 * gu, new_bubble_height - 6 * gu, 1.0, 0.0);

	cairo_destroy (cr);

	//update_shape (self);
}

void
bubble_set_synchronous (Bubble *self,
			const gchar *sync)
{
	BubblePrivate* priv;

	g_return_if_fail (IS_BUBBLE (self));

	priv = GET_PRIVATE (self);

	if (priv->synchronous != NULL)
		g_free (priv->synchronous);

	priv->synchronous = g_strdup (sync);
}

void
bubble_set_sender (Bubble *self,
		   const gchar *sender)
{
	BubblePrivate* priv;

	g_return_if_fail (IS_BUBBLE (self));

	priv = GET_PRIVATE (self);

	g_string_assign (priv->sender, sender);
}

gboolean
bubble_is_synchronous (Bubble *self)
{
	if (!self || !IS_BUBBLE (self))
		return FALSE;

	return (GET_PRIVATE (self)->synchronous != NULL);
}

gboolean
bubble_is_urgent (Bubble *self)
{
	g_return_val_if_fail (IS_BUBBLE (self), FALSE);

	return (GET_PRIVATE (self)->urgency == 2);
}

guint
bubble_get_urgency (Bubble *self)
{
	g_return_val_if_fail (IS_BUBBLE (self), 0);

	return GET_PRIVATE (self)->urgency;
}

void
bubble_set_urgency (Bubble *self,
		   guint  urgency)
{
	g_return_if_fail (IS_BUBBLE (self));

	GET_PRIVATE (self)->urgency = urgency;
}

void
bubble_determine_layout (Bubble* self)
{
	BubblePrivate* priv;

	/* sanity test */
	if (!self || !IS_BUBBLE (self))
		return;

	priv = GET_PRIVATE (self);

	/* set a sane default */
	priv->layout = LAYOUT_NONE;

	/* icon-only layout-case, e.g. eject */
	if (priv->icon_only && priv->icon_pixbuf != NULL)
	{
		priv->layout = LAYOUT_ICON_ONLY;
		return;
	}

	/* icon + indicator layout-case, e.g. volume */
	if ((priv->icon_pixbuf       != NULL) &&
	    (priv->title->len        != 0) &&
	    (priv->message_body->len == 0) &&
	    (priv->value             >= -1))
	{
		priv->layout = LAYOUT_ICON_INDICATOR;
		return;
	}

	/* icon + title layout-case, e.g. "Wifi signal lost" */
	if ((priv->icon_pixbuf       != NULL) &&
	    (priv->title->len        != 0) &&
	    (priv->message_body->len == 0) &&
	    (priv->value             == -2))
	{
		priv->layout = LAYOUT_ICON_TITLE;
		return;
	}

	/* icon/avatar + title + body/message layout-case, e.g. IM-message */
	if ((priv->icon_pixbuf       != NULL) &&
	    (priv->title->len        != 0) &&
	    (priv->message_body->len != 0) &&
	    (priv->value             == -2))
	{
		priv->layout = LAYOUT_ICON_TITLE_BODY;
		return;
	}

	/* title + body/message layout-case, e.g. IM-message without avatar */
	if ((priv->icon_pixbuf       == NULL) &&
	    (priv->title->len        != 0) &&
	    (priv->message_body->len != 0) &&
	    (priv->value             == -2))
	{
		priv->layout = LAYOUT_TITLE_BODY;
		return;
	}

	/* title-only layout-case, use discouraged but needs to be supported */
	if ((priv->icon_pixbuf       == NULL) &&
	    (priv->title->len        != 0) &&
	    (priv->message_body->len == 0) &&
	    (priv->value             == -2))
	{
		priv->layout = LAYOUT_TITLE_ONLY;
		return;
	}

	return;
}

BubbleLayout
bubble_get_layout (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return LAYOUT_NONE;

	return GET_PRIVATE (self)->layout;
}

void
bubble_set_icon_only (Bubble*  self,
		      gboolean allowed)
{
	if (!self || !IS_BUBBLE (self))
		return;

	GET_PRIVATE (self)->icon_only = allowed;
}

void
bubble_set_append (Bubble*  self,
		   gboolean allowed)
{
	if (!self || !IS_BUBBLE (self))
		return;

	GET_PRIVATE (self)->append = allowed;
}


gboolean
bubble_is_append_allowed (Bubble* self)
{
	if (!self || !IS_BUBBLE (self))
		return FALSE;

	return GET_PRIVATE (self)->append;
}

void
bubble_append_message_body (Bubble*      self,
			    const gchar* append_body)
{
	gboolean       result = FALSE;
	gchar*         text   = NULL;
	GError*        error  = NULL;
	BubblePrivate* priv   = NULL;

	if (!self || !IS_BUBBLE (self) || !append_body)
		return;

	priv = GET_PRIVATE (self);

	// filter out any HTML/markup if possible
    	result = pango_parse_markup (append_body,
				     -1,
				     0,    // no accel-marker needed
				     NULL, // no PangoAttr needed
				     &text,
				     NULL, // no accel-marker-return needed
				     &error);
	if (error && !result)
	{
		g_warning ("%s(): Got error \"%s\"\n",
			   G_STRFUNC,
			   error->message);
		g_error_free (error);
		error = NULL;

		if (text)
			g_free (text);
	}

	if (text)
	{
		// append text to current message-body
		g_string_append (priv->message_body, text);
		priv->message_body_needs_refresh = TRUE;

		g_signal_emit (self,
			       g_bubble_signals[MESSAGE_BODY_INSERTED],
			       0,
			       text);

		g_free (text);
	}
}

gdouble
align (gdouble  val,
       gboolean odd)
{
	gdouble fract = val - (gint) val;

	if (odd)
	{
		/* for strokes with an odd line-width */
		if (fract != 0.5)
		  return (gdouble) ((gint) val + 0.5);
		else
		  return val;
	}
	else
	{
		/* for strokes with an even line-width */
		if (fract != 0.0)
		  return (gdouble) ((gint) val);
		else
		  return val;
	}
}

void
setup_color (gdouble*  out,
             GdkColor* color,
             gdouble   opacity)
{
	out[0] = (gdouble) color->red / (gdouble) 0xFFFF;
	out[1] = (gdouble) color->green / (gdouble) 0xFFFF;
	out[2] = (gdouble) color->blue / (gdouble) 0xFFFF;
	out[3] = opacity;
}

void
setup_font_weight (PangoWeight* pweight,
                   FontWeight   fweight)
{
	*pweight = PANGO_WEIGHT_NORMAL;

	if (fweight == FONT_WEIGHT_LIGHT)
		*pweight = PANGO_WEIGHT_MEDIUM;
	else if (fweight == FONT_WEIGHT_NORMAL)
		*pweight = PANGO_WEIGHT_NORMAL;
	else if (fweight == FONT_WEIGHT_BOLD)
		*pweight = PANGO_WEIGHT_BOLD;
}

cairo_surface_t*
bubble_get_scratch_surface (Bubble* self)
{
	g_return_val_if_fail (IS_BUBBLE (self), NULL);

	return GET_PRIVATE (self)->scratch;
}

void
bubble_set_secondary_icon_from_pixbuf (Bubble*    self,
									   GdkPixbuf* pixbuf)
{
	g_return_if_fail (IS_BUBBLE (self) && !pixbuf);
	GET_PRIVATE (self)->secondary_icon_pixbuf = (GdkPixbuf*) g_object_ref (pixbuf);
	_refresh_secondary_icon (self);
}

GdkPixbuf*
bubble_get_secondary_icon_pixbuf (Bubble* self)
{
	g_return_val_if_fail (IS_BUBBLE (self), NULL);
	return GET_PRIVATE (self)->secondary_icon_pixbuf;
}

void
bubble_get_ignore_geometry (Bubble*        self,
							nux::Geometry* geo)
{
	g_return_if_fail (IS_BUBBLE (self));

	gint width  = 0;
	gint height = 0;
	gint gu     = GET_PRIVATE(self)->_units->getGU ();
	bubble_get_size (self, &width, &height);

	geo->x      = 2 * gu;
	geo->y      = 4 * gu + height - 6 * gu;
	geo->width  = 17 * gu;
	geo->height = 4 * gu;
}

void
bubble_get_accept_geometry (Bubble*        self,
							nux::Geometry* geo)
{
	g_return_if_fail (IS_BUBBLE (self));

	gint width  = 0;
	gint height = 0;
	gint gu     = GET_PRIVATE(self)->_units->getGU ();
	bubble_get_size (self, &width, &height);

	geo->x      = 21 * gu;
	geo->y      = 4 * gu + height - 6 * gu;
	geo->width  = 17 * gu;
	geo->height = 4 * gu;
}
