/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** notify-osd
**
** display.c - manages the display of notifications waiting in the stack/queue
**
** Copyright 2009 Canonical Ltd.
**
** Authors:
**    Mirco "MacSlow" Mueller <mirco.mueller@canonical.com>
**    David Barth <david.barth@canonical.com>
**
** This program is free software: you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 3, as published
** by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranties of
** MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
** PURPOSE.  See the GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program.  If not, see <http://www.gnu.org/licenses/>.
**
*******************************************************************************/

/* This is actually part of stack.c, but moved here because it manages
   the display of notifications. This is also in preparation for some
   refactoring, where we'll create:
 *  - a distinct (non-graphical) notification.c module
 *  - a distinct display.c module that tkaes care of placing bubbles
 *    on the screen
 */
  

static Bubble*
stack_find_bubble_on_display (Stack *self)
{
	GList*    list   = NULL;
	Bubble*   bubble = NULL;

	g_assert (IS_STACK (self));

	/* find the bubble on display */
	for (list = g_list_first (self->list);
	     list != NULL;
	     list = g_list_next (list))
	{
		bubble = (Bubble*) list->data;

		if (bubble_is_visible (bubble))
			return bubble;
	}

	return NULL;
}

static gboolean
stack_is_at_top_corner (Stack *self, Bubble *bubble)
{
	gint x, y1, y2;

	g_assert (IS_STACK (self));
	g_assert (IS_BUBBLE (bubble));

	defaults_get_top_corner (self->defaults, &x, &y1);
	bubble_get_position (bubble, &x, &y2);

	return y1 == y2;
}

static void
stack_display_position_sync_bubble (Stack *self, Bubble *bubble)
{
	gint y = 0;
	gint x = 0;

	x = defaults_get_desktop_right (self->defaults) + 64 -
		EM2PIXELS (defaults_get_bubble_width (self->defaults), self->defaults) -
		EM2PIXELS (defaults_get_bubble_horz_gap (self->defaults), self->defaults) -
		2 * EM2PIXELS (defaults_get_bubble_shadow_size (self->defaults), self->defaults);
	y = defaults_get_desktop_bottom (self->defaults) -
		EM2PIXELS (defaults_get_bubble_min_height (self->defaults), self->defaults);

	bubble_move (bubble, x, y);
}

static void

stack_display_sync_bubble (Stack *self, Bubble *bubble)
{
	g_return_if_fail (IS_STACK (self));
	g_return_if_fail (IS_BUBBLE (bubble));

	bubble_set_timeout (bubble, 2000);

	Bubble *other = stack_find_bubble_on_display (self);
	if (other != NULL)
		bubble_refresh (other);

	/* is the notification reusing the current bubble? */
	if (sync_bubble == bubble)
	{
		bubble_start_timer (bubble, TRUE);
		bubble_refresh (bubble);
		return;
	}

	stack_display_position_sync_bubble (self, bubble);

	sync_bubble = bubble;

	g_signal_connect (G_OBJECT (bubble),
			  "timed-out",
			  G_CALLBACK (close_handler),
			  self);
}

static Bubble*
stack_select_next_to_display (Stack *self)
{
	Bubble*   next_to_display = NULL;
	GList*    list   = NULL;
	Bubble*   bubble = NULL;

	/* pickup the next bubble to display */
	for (list = g_list_first (self->list);
	     list != NULL;
	     list = g_list_next (list))
	{
		bubble = (Bubble*) list->data;

		/* sync. bubbles have already been taken care of */
		if (bubble_is_synchronous (bubble))
		{
			g_critical ("synchronous notifications are managed separately");
			continue;
		}

		/* if there is already one bubble on display
		   we don't have room for another one */
		if (bubble_is_visible (bubble))
			return NULL;

		if (bubble_is_urgent (bubble))
		{
			/* pick-up the /first/ urgent bubble
			   in the queue (FIFO) */
			return bubble;

		}

		if (next_to_display == NULL)
			next_to_display = bubble;

		/* loop, in case there are urgent bubbles waiting higher up
		   in the stack */
	}

	return next_to_display;
}

void
stack_layout (Stack* self)
{
	g_return_if_fail (self != NULL);

	Notification* sd      = NULL;
	Notification* regular = NULL;

	if (queue_get_size (self->sd_queue) != 0)
		sd = queue_pop (self->sd_queue);

	if (sd)
	{
		gchar* switch_to_app_name = notification_get_switch_to_application(sd);
		self->view->SetAppSwitchName(switch_to_app_name);

		Bubble* bubble = bubble_new (self->defaults);
		bubble_set_notification (bubble, sd);
		bubble_determine_layout (bubble);
		bubble_recalc_size (bubble);
		self->view->SetSourceSurface (bubble_get_scratch_surface (bubble));
		/*bubble_set_timeout (bubble, notification_get_onscreen_time (sd));
		bubble_start_timer (bubble, FALSE);*/

		// set hit-areas for snap-decision buttons
		nux::Geometry ignore_geo;
		nux::Geometry accept_geo;
		bubble_get_ignore_geometry (bubble, &ignore_geo);
		bubble_get_accept_geometry (bubble, &accept_geo);
		self->view->SetIgnoreGeometry (ignore_geo);
		self->view->SetAcceptGeometry (accept_geo);

		// set actions for snap-decision related button-callbacks
		g_string_assign (self->sd_dbus_sender, notification_get_dbus_sender (sd));
		self->sd_id = notification_get_id (sd);
		self->regular_bubble = NULL;
		g_string_assign (self->sd_ignore_action, notification_get_action_by_index (sd, 2));
		g_string_assign (self->sd_accept_action, notification_get_action_by_index (sd, 0));

		/*g_signal_connect (G_OBJECT(bubble),
						  "timed-out",
						  G_CALLBACK(close_handler),
						  self);*/
		self->view->Show();
	}
	else
	{
		if (queue_get_size (self->ep_queue) != 0)
			regular = queue_pop (self->ep_queue);

		if (regular)
		{
			gchar* switch_to_app_name = notification_get_switch_to_application(regular);
			self->view->SetAppSwitchName(switch_to_app_name);

			Bubble* bubble = bubble_new (self->defaults);
			bubble_set_notification (bubble, regular);
			bubble_determine_layout (bubble);
			bubble_recalc_size (bubble);
			self->view->SetSourceSurface (bubble_get_scratch_surface (bubble));
			bubble_set_timeout (bubble, defaults_get_on_screen_timeout (self->defaults));
			bubble_start_timer (bubble, FALSE);

			// reset any hit-area, since we're not a snap-decision
			nux::Geometry geo = {0, 0, 0, 0};
			self->view->SetIgnoreGeometry (geo);
			self->view->SetAcceptGeometry (geo);

			// reset snap-decision related values
			g_string_assign (self->sd_dbus_sender, "");
			self->sd_id = 0;
			self->regular_bubble = bubble;
			g_string_assign (self->sd_ignore_action, "");
			g_string_assign (self->sd_accept_action, "");

			g_signal_connect (G_OBJECT(bubble),
							  "timed-out",
							  G_CALLBACK(close_handler),
							  self);
			self->view->Show();

			queue_remove_by_id (self->ep_queue, notification_get_id (regular));
		}		
	}
}
