/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** notify-osd
**
** stack.c - manages the stack/queue of incoming notifications
**
** Copyright 2009 Canonical Ltd.
**
** Authors:
**    Mirco "MacSlow" Mueller <mirco.mueller@canonical.com>
**    David Barth <david.barth@canonical.com>
**
** Contributor(s):
**    Abhishek Mukherjee <abhishek.mukher.g@gmail.com> (append fixes, rev. 280)
**    Aurélien Gâteau <aurelien.gateau@canonical.com> (0.10 spec, rev. 348)
**
** This program is free software: you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 3, as published
** by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranties of
** MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
** PURPOSE.  See the GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program.  If not, see <http://www.gnu.org/licenses/>.
**
*******************************************************************************/

#include <Nux/Nux.h>
#include <Nux/VLayout.h>
#include <Nux/WindowThread.h>

#include <assert.h>
#include "dbus.h"
#include <dbus/dbus-glib-lowlevel.h>
#include <gtk/gtk.h>
#include <glib-object.h>
#include "stack.h"
#include "bubble.h"
#include "apport.h"
#include "dialog.h"
#include "dnd.h"
#include "log.h"
#include "notifications-view.h"
#include "units.h"

G_DEFINE_TYPE (Stack, stack, G_TYPE_OBJECT);

#define FORCED_SHUTDOWN_THRESHOLD 500
#define STACK_LAYOUT_TRIGGER_PAUSE 500

/* fwd declaration */
void close_handler (GObject* n, Stack* stack);

/*-- internal API ------------------------------------------------------------*/

static void
stack_dispose (GObject* gobject)
{
	/* chain up to the parent class */
	G_OBJECT_CLASS (stack_parent_class)->dispose (gobject);
}

static
void
disconnect_bubble (gpointer data,
	      gpointer user_data)
{
	Bubble* bubble = BUBBLE(data);
	Stack* stack = STACK(user_data);
	g_signal_handlers_disconnect_by_func (G_OBJECT (bubble),
										  (gpointer) close_handler,
										  stack);
}

static void
stack_finalize (GObject* gobject)
{
	if (STACK(gobject)->list != NULL)
		g_list_foreach (STACK(gobject)->list, disconnect_bubble, gobject);
	if (STACK(gobject)->defaults != NULL)
		g_object_unref (STACK(gobject)->defaults);
	if (STACK(gobject)->ep_queue != NULL)
		queue_del (STACK(gobject)->ep_queue);
	if (STACK(gobject)->sd_queue != NULL)
		queue_del (STACK(gobject)->sd_queue);
	if (STACK(gobject)->visual_queue != NULL)
		visual_queue_del (STACK(gobject)->visual_queue);
	if (STACK(gobject)->sd_dbus_sender != NULL)
		g_string_free (STACK(gobject)->sd_dbus_sender, true);
	if (STACK(gobject)->sd_ignore_action != NULL)
		g_string_free (STACK(gobject)->sd_ignore_action, true);
	if (STACK(gobject)->sd_accept_action != NULL)
		g_string_free (STACK(gobject)->sd_accept_action, true);

	/* chain up to the parent class */
	G_OBJECT_CLASS (stack_parent_class)->finalize (gobject);
}

static void
stack_init (Stack* self)
{
	/* If you need specific construction properties to complete
	** initialization, delay initialization completion until the
	** property is set. */

	self->list = NULL;
	self->wt   = NULL;
}

static void
stack_get_property (GObject*    gobject,
		    guint       prop,
		    GValue*     value,
		    GParamSpec* spec)
{
	switch (prop)
	{
		default :
			G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop, spec);
		break;
	}
}

#include "stack-glue.h"

static void
stack_class_init (StackClass* klass)
{
	GObjectClass* gobject_class = G_OBJECT_CLASS (klass);

	gobject_class->dispose      = stack_dispose;
	gobject_class->finalize     = stack_finalize;
	gobject_class->get_property = stack_get_property;

	dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass),
					 &dbus_glib_stack_object_info);
}

gint
compare_id (gconstpointer a,
	    gconstpointer b)
{
	gint  result;
	guint id_1;
	guint id_2;

	if (!a || !b)
		return -1;

	if (!IS_BUBBLE (a))
		return -1;

	id_1 = bubble_get_id ((Bubble*) a);
	id_2 = *((guint*) b);

	if (id_1 < id_2)
		result = -1;

	if (id_1 == id_2)
		result = 0;

	if (id_1 > id_2)
		result = 1;

	return result;
}

/*static gint
compare_append (gconstpointer a,
	        gconstpointer b)
{
	const gchar* str_1;
	const gchar* str_2;
	gint cmp;

	if (!a || !b)
		return -1;

	if (!IS_BUBBLE (a))
		return -1;
	if (!IS_BUBBLE (b))
		return 1;

	if (!bubble_is_append_allowed((Bubble*) a))
		return -1;
	if (!bubble_is_append_allowed((Bubble*) b))
		return 1;

	str_1 = bubble_get_title ((Bubble*) a);
	str_2 = bubble_get_title ((Bubble*) b);

	cmp = g_strcmp0(str_1, str_2);
	if (cmp < 0)
		return -1;
	if (cmp > 0)
		return 1;

	str_1 = bubble_get_sender ((Bubble*) a);
	str_2 = bubble_get_sender ((Bubble*) b);
	return g_strcmp0(str_1, str_2);
}*/

GList*
find_entry_by_id (Stack* self,
		  guint  id)
{
	GList* entry;

	/* sanity check */
	if (!self)
		return NULL;

	entry = g_list_find_custom (self->list,
				    (gconstpointer) &id,
				    compare_id);
	if (!entry)
		return NULL;

	return entry;
}

static Bubble*
find_bubble_by_id (Stack* self,
		   guint  id)
{
	GList* entry;

	/* sanity check */
	if (!self)
		return NULL;

	entry = g_list_find_custom (self->list,
				    (gconstpointer) &id,
				    compare_id);
	if (!entry)
		return NULL;

	return (Bubble*) entry->data;
}

/*static Bubble*
find_bubble_for_append(Stack* self,
		       Bubble *bubble)
{
	GList* entry;

	* sanity check *
	if (!self)
		return NULL;

	entry = g_list_find_custom(self->list,
				   (gconstpointer) bubble,
				   compare_append);

	if (!entry)
		return NULL;

	return (Bubble*) entry->data;
}*/

/*static void
_weak_notify_cb (gpointer data,
		 GObject* former_object)
{
	Stack* stack = STACK (data);

	stack->list = g_list_remove (stack->list, former_object);
}*/

static void
_trigger_bubble_redraw (gpointer data,
			gpointer user_data)
{
	Bubble* bubble;

	if (!data)
		return;

	bubble = BUBBLE (data);
	if (!IS_BUBBLE (bubble))
		return;

	bubble_recalc_size (bubble);
	bubble_refresh (bubble);
}

static void
value_changed_handler (Defaults* defaults,
		       Stack*    stack)
{
	if (stack->list != NULL)
		g_list_foreach (stack->list, _trigger_bubble_redraw, NULL);
}

static Bubble *sync_bubble = NULL;

#include "display.cpp"

void
visual_queue_empty_handler (VisualQueue* visual_queue,
                            gpointer     data)
{
	Stack* stack = STACK (data);

	stack_layout (stack);
}

/*-- public API --------------------------------------------------------------*/

Stack*
stack_new (Defaults* defaults)
{
	Stack* self;

	if (!defaults)
		return NULL;

	self = (Stack*) g_object_new (STACK_TYPE, NULL);
	if (!self)
		return NULL;

	self->defaults           = defaults;
	self->list               = NULL;
	self->next_id            = 1;
	self->slots[SLOT_TOP]    = NULL;
	self->slots[SLOT_BOTTOM] = NULL;
	self->ep_queue           = queue_new ();
	self->sd_queue           = queue_new ();
	self->visual_queue       = visual_queue_new (defaults_get_desktop_right (defaults) - EM2PIXELS (20.0, defaults),
	                                             defaults_get_desktop_top (defaults) + EM2PIXELS (0.5, defaults),
	                                             defaults_get_desktop_bottom (defaults),
	                                             self->sd_queue);
	self->sd_dbus_sender     = g_string_new ("");
	self->sd_id              = 0;
	self->sd_ignore_action   = g_string_new ("");
	self->sd_accept_action   = g_string_new ("");
	self->_units             = Units::getInstance ();

	self->model = dee_shared_model_new (MODEL_NAME);
	dee_model_set_schema (self->model,
                          "i",    /* unique notification-id */
                          "s",    /* summary-text strdup'ed from setter */
                          "s",    /* body-text strdup'ed from setter */
                          "i",    /* range: 0..100, special: -1, 101 */
                          "(iv)", /* icon-flag tuple for icon */
                          "i",    /* time on-screen in ms */
                          "s",    /* DBus sender-name */
                          "s",    /* app-name, strdup'ed from setter */
                          "i",    /* pid of sending application */
                          "u",    /* timestamp of reception - tv_sec part  */
                          "u",    /* timestamp of reception - tv_usec part */
                          "u",    /* urgency-level: low, normal, high */
                          //"as",   /* list of labels for actions/buttons */
                          //"u",    /* number of labels */
                          "b",    /* flag to indicate, that it's shown to user */
                          "b",    /* flag to indicate if/if not to use tint */
                          "b",    /* flag to indicate append use-case */
                          "b",    /* flag to mark sychronous vs. asychronous  */
                          "b",    /* flag to indicate icon-only use-case */
                          NULL);
    g_assert (DEE_IS_MODEL (self->model));

	/* hook up handler to act on changes of defaults/settings */
	g_signal_connect (G_OBJECT (defaults),
			  "value-changed",
			  G_CALLBACK (value_changed_handler),
			  self);

	/* a way to tell stack that it can continue to process asynchronous
	** notifications*/
	g_signal_connect (G_OBJECT (self->visual_queue),
	                  "empty",
	                  G_CALLBACK (visual_queue_empty_handler),
	                  self);

	self->initial_bubble = true;
	self->regular_bubble = NULL;
	self->view = new NotificationsView();		
	self->view->FadedOut.connect (bind(sigc::ptr_fun(stack_mark_queue_empty), self));
	self->view->mouse_up.connect([self] (int x, int y, unsigned long foo, unsigned long bar)
	{
		if ((y < 4 * self->_units->getGU ()) ||
			(y > self->view->height + 4 * self->_units->getGU ()))
			return;

		if (self->view->GetAppSwitchName() && self->view->GetVisible())
		{
			std::string name (self->view->GetAppSwitchName());
			g_print("switch to app: %s\n", name.c_str());
			nux::GetWindowThread()->GetGraphicsDisplay().SwitchToApplication(name);
			self->view->Hide ();
		}

		/* dbus-sender, id, action-string[2], action-string[0] */
		if (self->view->IsIgnoreHit(x, y))
		{
			dbus_send_action_signal (self->sd_dbus_sender->str,
				                     self->sd_id,
				                     self->sd_ignore_action->str);
			dbus_send_close_signal (self->sd_dbus_sender->str, self->sd_id, 3);

			queue_remove_by_id (self->sd_queue, self->sd_id);
			self->view->Hide ();
		}


		if (self->view->IsAcceptHit(x, y))
		{
			dbus_send_action_signal (self->sd_dbus_sender->str,
			                         self->sd_id,
			                         self->sd_accept_action->str);
			dbus_send_close_signal (self->sd_dbus_sender->str, self->sd_id, 3);

			queue_remove_by_id (self->sd_queue, self->sd_id);
			self->view->Hide ();
		}
	});
	self->layout = new nux::VLayout(NUX_TRACKER_LOCATION);
	self->layout->AddView(self->view, 1, nux::eCenter, nux::eFull);
	self->layout->SetContentDistribution(nux::eStackCenter);
	nux::GetWindowThread ()->SetLayout (self->layout);		

	return self;
}

void
stack_del (Stack* self)
{
	if (!self)
		return;

	g_object_unref (self);
}

gboolean
trigger_stack_layout (gpointer data)
{
	Stack* stack = (Stack*) data;
	stack_layout (stack);
	return false;
}

void
close_handler (GObject* bubble,
	       Stack*  stack)
{
	stack->view->Hide();

	guint id = bubble_get_id (BUBBLE (bubble));
	Notification* n = NULL;
	n = queue_find_by_id (stack->sd_queue, (gint) id);
	if (!n)
		n = queue_find_by_id (stack->ep_queue, (gint) id);

	if (IS_NOTIFICATION (n))
	{
		dbus_send_close_signal (notification_get_dbus_sender (n), id, 3);
		if (notification_get_num_actions (n) > 0)
			queue_remove_by_id (stack->sd_queue, id);
		else
			queue_remove_by_id (stack->ep_queue, id);
	}

	// add this "break" to allow the fade-out animation of Hide() to finish
	g_timeout_add (STACK_LAYOUT_TRIGGER_PAUSE, trigger_stack_layout, stack);

	return;
}

/* since notification-ids are unsigned integers the first id index is 1, 0 is
** used to indicate an error */
guint
stack_push_bubble (Stack*  self,
		   Bubble* bubble)
{
	guint notification_id = -1;

	/* sanity check */
	if (!self || !IS_BUBBLE (bubble))
		return -1;

	/* check if this is just an update */
	if (find_bubble_by_id (self, bubble_get_id (bubble)))
	{
		bubble_start_timer (bubble, TRUE);
		bubble_refresh (bubble);

		return bubble_get_id (bubble);
	}

	/* add bubble/id to stack */
	notification_id = self->next_id++;

	// FIXME: migrate stack to use abstract notification object and don't
	// keep heavy bubble objects around, at anyone time at max. only two
	// bubble-objects will be in memory... this will also reduce leak-
	// potential
	bubble_set_id (bubble, notification_id);
	self->list = g_list_append (self->list, (gpointer) bubble);

	g_signal_connect (G_OBJECT (bubble),
			  "timed-out",
			  G_CALLBACK (close_handler),
			  self);

	/* return current/new id to caller (usually our DBus-dispatcher) */
	return notification_id;
}

void
stack_pop_bubble_by_id (Stack* self,
			guint  id)
{
	Bubble* bubble;

	/* sanity check */
	if (!self)
		return;

	/* find bubble corresponding to id */
	bubble = find_bubble_by_id (self, id);
	if (!bubble)
		return;

	/* find entry in list corresponding to id and remove it */
	self->list = g_list_delete_link (self->list,
					 find_entry_by_id (self, id));
	g_object_unref (bubble);

	/* immediately refresh the layout of the stack */
	stack_layout (self);
}

static GdkPixbuf*
process_dbus_icon_data (GValue *data)
{
	GType dbus_icon_t;
	GArray *pixels;
	int width, height, rowstride, bits_per_sample, n_channels, size;
	gboolean has_alpha;
	guchar *copy;
	GdkPixbuf *pixbuf = NULL;

	g_return_val_if_fail (data != NULL, NULL);

	dbus_icon_t = dbus_g_type_get_struct ("GValueArray",
					      G_TYPE_INT,
					      G_TYPE_INT,
					      G_TYPE_INT,
					      G_TYPE_BOOLEAN,
					      G_TYPE_INT,
					      G_TYPE_INT,
					      dbus_g_type_get_collection ("GArray",
									  G_TYPE_UCHAR),
					      G_TYPE_INVALID);
	
	if (G_VALUE_HOLDS (data, dbus_icon_t))
	{
		dbus_g_type_struct_get (data,
					0, &width,
					1, &height,
					2, &rowstride,
					3, &has_alpha,
					4, &bits_per_sample,
					5, &n_channels,
					6, &pixels,
					G_MAXUINT);

		size = (height - 1) * rowstride + width *
			((n_channels * bits_per_sample + 7) / 8);
		copy = (guchar *) g_memdup (pixels->data, size);
		pixbuf = gdk_pixbuf_new_from_data(copy, GDK_COLORSPACE_RGB,
						  has_alpha,
						  bits_per_sample,
						  width, height,
						  rowstride,
						  (GdkPixbufDestroyNotify)g_free,
						  NULL);
	}

	return pixbuf;
}

// control if there are non-default actions requested with this notification
static gboolean
dialog_check_timeout (gint timeout)
{
	gboolean turn_into_dialog = FALSE;

	if (timeout == 0)
	{
		turn_into_dialog = TRUE;
		g_debug ("notification request turned into a dialog "
			 "box, because of its infinite timeout");
	}

	return turn_into_dialog;
}

static gboolean
check_for_actions (gchar** actions)
{
	gint     i                     = 0;
	gboolean uses_snap_decision    = FALSE;
	gboolean uses_only_one_action  = FALSE;
	gboolean uses_too_many_actions = FALSE;
	gboolean uses_empty_labels     = FALSE;

	if (actions != NULL)
	{
		for (i = 0; actions[i] != NULL; i += 2)
		{
			if (actions[i+1] == NULL)
			{
				g_debug ("incorrect action callback with no label\n");
				uses_empty_labels = TRUE;
				break;
			}

			if (i > 12)
			{
				g_debug ("uses more than six actions\n");
				uses_too_many_actions = TRUE;
				break;
			}
		}

		if (i == 4 || i == 8 || i == 10 || i == 12)
		{
			g_debug ("uses correct amount of actions\n");
		}

		if (i == 2)
		{
			g_debug ("uses just one action\n");
			uses_only_one_action = TRUE;
		}

		if (uses_only_one_action  ||
		    uses_too_many_actions ||
		    uses_empty_labels     ||
		    i == 0)
			uses_snap_decision = FALSE;
		else
			uses_snap_decision = TRUE;
	}

	return uses_snap_decision;
}

// FIXME: a intnernal function used for the forcefully-shutdown work-around
// regarding mem-leaks
gboolean
_arm_forced_quit (gpointer data)
{
	Stack* stack = NULL;

	// sanity check, "disarming" this forced quit
	if (!data)
		return FALSE;

	stack = STACK (data);

	// only forcefully quit if the queue is empty
	if (g_list_length (stack->list) == 0)
	{
		gtk_main_quit ();

		// I don't think this is ever reached :)
		return FALSE;
	}

	return TRUE;
}

void
_remove_slot_allocation_cb (Bubble*  bubble,
                            gpointer data)
{
	Stack* stack = STACK (data);

	stack_free_slot (stack, bubble);
}

guint
stack_notify_handler (Stack*                 self,
                      const gchar*           app_name,
                      guint                  id,
                      const gchar*           icon,
                      const gchar*           summary,
                      const gchar*           body,
                      gchar**                actions,
                      GHashTable*            hints,
                      gint                   timeout,
                      DBusGMethodInvocation* context)
{
	GValue*       data               = NULL;
	GValue*       compat             = NULL;
	GdkPixbuf*    pixbuf             = NULL;
	gboolean      turn_into_dialog   = FALSE;
	gboolean      uses_snap_decision = FALSE;
	gint          notification_id;
	Notification* n                  = NULL;

	if (id != 0)
		notification_id = id;
	else
		notification_id = self->next_id++;

	// check max. allowed limit queue-size
	if (g_list_length (self->list) > MAX_STACK_SIZE)
	{
		GError* error = NULL;

		error = g_error_new (g_quark_from_string ("notify-osd"),
		                     1,
		                     "Reached stack-limit of %d",
		                     MAX_STACK_SIZE);
		dbus_g_method_return_error (context, error);
		g_error_free (error);
		error = NULL;

		return TRUE;
	}

	// see if pathological timeouts are used by an app issuing a notification
	turn_into_dialog = FALSE;
	if (turn_into_dialog)
	{
		// TODO: apport_report (app_name, summary, actions, timeout);
		gchar* sender = dbus_g_method_get_sender (context);

		fallback_dialog_show (self->defaults,
				      sender,
				      app_name,
				      notification_id,
				      summary,
				      body,
				      actions);
		g_free (sender);
		dbus_g_method_return (context, notification_id);

		return TRUE;
	}

	// check if the calling application wants to use snap-decision/action
	if (hints)
	{
		data = (GValue*) g_hash_table_lookup (hints, "x-canonical-snap-decisions");
		
		if (data && G_VALUE_HOLDS_STRING (data))
		{
			// check if calling application uses correct number of actions
			uses_snap_decision = check_for_actions (actions);
			if (uses_snap_decision)
			{
				/* FIXME: unable to obtain the PID of the process which
				** triggered the notification, so the DBus sender-name is used..
				** this is equally good for the moment 
				**
				** - with this we can protect against flooding from a single app
				** and throttle it to one snap-decision at a time per app
				**
				** - what the DBus sender-name does not allow us to do, is to
				** remove notifications in the queue - prior to being shown - if
				** the sending app has been quit in the meantime */

				n = notification_new ();
				notification_set_id (n, notification_id);
				notification_set_summary (n, summary);
				notification_set_body (n, body);
				notification_set_onscreen_time (n, 30 * 1000); // 30 secs.

				/* try to load the icon in any of the four known ways */
				if (hints)
				{
					if ((data = (GValue*) g_hash_table_lookup (hints, "image_data")))
					{
						g_debug("Using image_data hint\n");
						pixbuf = process_dbus_icon_data (data);
						notification_set_icon_from_pixbuf (n, pixbuf);
					}
					else if ((data = (GValue*) g_hash_table_lookup (hints, "image_path")))
					{
						g_debug("Using image_path hint\n");
						if ((data && G_VALUE_HOLDS_STRING (data)))
							notification_set_icon_from_path (n, g_value_get_string(data));
						else
							g_warning ("image_path hint is not a string\n");
					}
					else if (icon && *icon != '\0')
					{
						g_debug("Using icon parameter\n");
						notification_set_icon (n, icon);
					}
					else if ((data = (GValue*) g_hash_table_lookup (hints, "icon_data")))
					{
						g_debug("Using deprecated icon_data hint\n");
						pixbuf = process_dbus_icon_data (data);
						notification_set_icon_from_pixbuf (n, pixbuf);
					}
				}

				/*notification_set_onscreen_time (n, const gint screen_time);*/
				gchar* dbus_sender = dbus_g_method_get_sender (context);
				notification_set_dbus_sender (n, dbus_sender);
				g_free (dbus_sender);
				notification_set_sender_name (n, app_name);

				if (hints)
				{
					data = (GValue*) g_hash_table_lookup (
										hints,
					                    "x-canonical-private-button-tint");
			
					if (data && G_VALUE_HOLDS_STRING (data))
						notification_set_button_tint (n, TRUE);
					else
						notification_set_button_tint (n, FALSE);
				}

				if (hints)
				{
					data = (GValue*) g_hash_table_lookup (hints, "urgency");
					if (data && G_VALUE_HOLDS_UCHAR (data))
						notification_set_urgency (n, (Urgency) g_value_get_uchar (data));
				}
				notification_set_actions (n, actions);
				queue_inject (self->sd_queue, n);

				//handle app-switch?
				if (hints)
				{
					data = (GValue*) g_hash_table_lookup(hints, "x-canonical-switch-to-application");
					if (data && G_VALUE_HOLDS_STRING(data))
					{
						const gchar* appname = g_value_get_string(data);
						notification_set_switch_to_application(n, appname);
					}
				}

				// handle secondary-icon hint
				if (hints)
				{
					data = (GValue*) g_hash_table_lookup(hints, "x-canonical-secondary-icon");
					if (data && G_VALUE_HOLDS_STRING(data))
						notification_set_secondary_icon_from_path (n, g_value_get_string(data));
				}

				// TODO: apport_report (app_name, summary, actions, timeout);
				dbus_g_method_return (context, notification_get_id (n));

				log_notification (n);

				// FIXME: pop any sd notification directly and pass it
				// over via libdee
				//Notification* regular = NULL;
				//regular = queue_pop (self->sd_queue);
				//stack_push_to_model (self, regular);

				if (self->initial_bubble)
				{
					self->initial_bubble = false;
					stack_layout (self);		
				}
				else if (self->regular_bubble != NULL)
				{
					// close any open non-SD bubble as a SD-bubble is always meant to immediately replace a non-SD one
					close_handler (G_OBJECT (self->regular_bubble), self);
					g_source_remove (bubble_get_timer_id (self->regular_bubble));
					self->regular_bubble = NULL;
				}

				return notification_id;
			}
		}
	}

	// deal with regular sync/async notifications
	n = NULL;

	/* handle append-hint */
	if (hints)
	{
		data   = (GValue*) g_hash_table_lookup (hints, "x-canonical-append");
		compat = (GValue*) g_hash_table_lookup (hints, "append");

		if ((data && G_VALUE_HOLDS_STRING (data)) ||
		    (compat && G_VALUE_HOLDS_STRING (compat)))
		{
			n = (Notification*) queue_find_by_summary (self->ep_queue, (gchar*) summary);
			if (n == NULL)
			{
				n = notification_new ();
				notification_set_id (n, notification_id);

				gchar* sender = dbus_g_method_get_sender (context);
				notification_set_sender_name (n, sender);
				g_free (sender);

				if (summary)
					notification_set_summary (n, summary);
			}

			notification_set_append_hint (n, TRUE);

			if (body)
			{
				GString* new_body = g_string_new (notification_get_body (n));
				g_string_append (new_body, "\n");
				g_string_append (new_body, body);
				notification_set_body (n, new_body->str);
				g_string_free (new_body, TRUE);
			}
		}
		else
		{
			n = queue_find_by_id (self->ep_queue, notification_id);
			if (n == NULL)
			{
				n = notification_new ();
				notification_set_id (n, notification_id);

				gchar* sender = dbus_g_method_get_sender (context);
				notification_set_sender_name (n, sender);
				g_free (sender);
			}

			/* set regular summary- and body-text */
			if (summary)
				notification_set_summary (n, summary);
			if (body)
				notification_set_body (n, body);

			notification_set_append_hint (n, FALSE);
		}
	}

	/* handle synchronous-hint */
	if (hints)
	{
		data   = (GValue*) g_hash_table_lookup (hints, "x-canonical-private-synchronous");
		//compat = (GValue*) g_hash_table_lookup (hints, "synchronous");
		if ((data && G_VALUE_HOLDS_STRING (data)) || (compat && G_VALUE_HOLDS_STRING (compat)))
		{
			if ((data && G_VALUE_HOLDS_STRING (data)) ||
			    (compat && G_VALUE_HOLDS_STRING (compat)))
				notification_set_sync_hint (n, TRUE);
			else
				notification_set_sync_hint (n, FALSE);
		}
	}

	/* handle value-hint */
	if (hints)
	{
		data = (GValue*) g_hash_table_lookup (hints, "value");
		if (data && G_VALUE_HOLDS_INT (data))
			notification_set_value (n, g_value_get_int (data));
	}

	/* handle urgency-hint */
	if (hints)
	{
		data = (GValue*) g_hash_table_lookup (hints, "urgency");
		if (data && G_VALUE_HOLDS_UCHAR (data))
			notification_set_urgency (n, (Urgency) g_value_get_uchar (data));
		else
			notification_set_urgency (n, URGENCY_LOW);
		/* Note: urgency was defined as an enum: LOW, NORMAL, CRITICAL
		** So, 2 means CRITICAL */
	}

	/* handle icon-only-hint */
	if (hints)
	{
		data   = (GValue*) g_hash_table_lookup (hints, "x-canonical-private-icon-only");
		compat = (GValue*) g_hash_table_lookup (hints, "icon-only");
		if ((data && G_VALUE_HOLDS_STRING (data)) || (compat && G_VALUE_HOLDS_STRING (compat)))
			notification_set_icon_only_hint (n, TRUE);
		else
			notification_set_icon_only_hint (n, FALSE);
	}

	/* handle all icon-data cases */
	if (hints)
	{
		if ((data = (GValue*) g_hash_table_lookup (hints, "image_data")))
		{
			g_debug("Using image_data hint\n");
			pixbuf = process_dbus_icon_data (data);
			notification_set_icon_from_pixbuf (n, pixbuf);
		}
		else if ((data = (GValue*) g_hash_table_lookup (hints, "image_path")))
		{
			g_debug("Using image_path hint\n");
			if ((data && G_VALUE_HOLDS_STRING (data)))
				notification_set_icon_from_path (n, g_value_get_string(data));
			else
				g_warning ("image_path hint is not a string\n");
		}
		else if (icon && *icon != '\0')
		{
			g_debug("Using icon parameter\n");
			notification_set_icon (n, icon);
		}
		else if ((data = (GValue*) g_hash_table_lookup (hints, "icon_data")))
		{
			g_debug("Using deprecated icon_data hint\n");
			pixbuf = process_dbus_icon_data (data);
			notification_set_icon_from_pixbuf (n, pixbuf);
		}
	}
  
	//handle app-switch?
	if (hints)
	{
		data = (GValue*) g_hash_table_lookup(hints, "x-canonical-switch-to-application");
		if (data && G_VALUE_HOLDS_STRING(data))
		{
			const gchar* appname = g_value_get_string(data);
			notification_set_switch_to_application(n, appname);
		}
	}

	// handle secondary-icon hint
	if (hints)
	{
		data = (GValue*) g_hash_table_lookup(hints, "x-canonical-secondary-icon");
		if (data && G_VALUE_HOLDS_STRING(data))
			notification_set_secondary_icon_from_path (n, g_value_get_string(data));
	}

	log_notification (n);

	queue_inject (self->ep_queue, n);

	dbus_g_method_return (context, notification_id);

	if (self->initial_bubble)
	{
		self->initial_bubble = false;
		stack_layout (self);		
	}

	return notification_id;
}

gboolean
stack_push_to_model (Stack*        self,
                     Notification* n)
{
	//GVariantBuilder builder_actions;
	//GVariant*       actions_variant = NULL;
	GVariant*       icon_variant = NULL;
	GTimeVal*       tv;

	/*g_variant_builder_init (&builder_actions, G_VARIANT_TYPE_ARRAY);
	if (notification_get_num_actions (n) == 0)
	{
		g_variant_builder_add_value (&builder_actions, "(not set)");
	}
	else
	{
		for (guint i = 0; i < notification_get_num_actions (n); i++)
			g_variant_builder_add_value (&builder_actions,
										 notification_get_action_by_index (n, i));
	}
	actions_variant = g_variant_builder_end (&builder_actions);*/

  // FIXME: using a place-holder icon for the time being
  icon_variant = g_variant_new ("(iv)", 0, g_variant_new_string ("distributor-logo"));

  tv = notification_get_timestamp (n);

  GString* dbus_sender = g_string_new (notification_get_dbus_sender (n));
  GString* sender_name = g_string_new (notification_get_sender_name (n));

  if (dbus_sender->str == NULL)
  	g_string_assign (dbus_sender, "(not set)");

  if (sender_name->str == NULL)
  	g_string_assign (sender_name, "(not set)");

  dee_model_append (self->model,
                    notification_get_id (n),
                    notification_get_summary (n),
                    notification_get_body (n),
                    notification_get_value (n),
                    icon_variant,
                    notification_get_onscreen_time (n),
                    dbus_sender->str,
                    sender_name->str,
                    notification_get_sender_pid (n),
                    tv->tv_sec,
                    tv->tv_usec,
                    notification_get_urgency (n),
                    //actions_variant,
                    //notification_get_num_actions (n),
                    notification_is_visible (n),
                    notification_get_button_tint (n),
                    notification_get_append_hint (n),
                    notification_get_sync_hint (n),
                    notification_get_icon_only_hint (n));

	g_string_free (dbus_sender, TRUE);
	g_string_free (sender_name, TRUE);

	return TRUE;
}

gboolean
stack_close_notification_handler (Stack*   self,
                                  guint    id,
                                  GError** error)
{
	if (id == 0)
		g_warning ("%s(): notification id == 0, likely wrong\n",
			   G_STRFUNC);
	else
	{
		Notification* n = queue_find_by_id (self->sd_queue, (gint) id);
		if (n)
		{
			dbus_send_close_signal (notification_get_dbus_sender (n), id, 3);
			queue_remove_by_id (self->sd_queue, (gint) id);
			self->view->Hide();
			stack_layout (self);
			return TRUE;
		}
	}

	Bubble* bubble = find_bubble_by_id (self, id);

	// exit but pretend it's ok, for applications
	// that call us after an action button was clicked
	if (bubble == NULL)
		return TRUE;

	dbus_send_close_signal (bubble_get_sender (bubble),
				bubble_get_id (bubble),
				3);

	stack_layout (self);

	return TRUE;
}

gboolean
stack_get_capabilities (Stack*   self,
                        gchar*** out_caps)
{
	*out_caps = (gchar**) g_malloc0 (18 * sizeof(char *));

	(*out_caps)[0]  = g_strdup ("actions");
	(*out_caps)[1]  = g_strdup ("body");
	(*out_caps)[2]  = g_strdup ("body-markup");
	(*out_caps)[3]  = g_strdup ("icon-static");
	(*out_caps)[4]  = g_strdup ("image/svg+xml");
	(*out_caps)[5]  = g_strdup ("x-canonical-private-synchronous");
	(*out_caps)[6]  = g_strdup ("x-canonical-append");
	(*out_caps)[7]  = g_strdup ("x-canonical-private-icon-only");
	(*out_caps)[8]  = g_strdup ("x-canonical-truncation");
	(*out_caps)[9]  = g_strdup ("x-canonical-snap-decisions");
	(*out_caps)[10] = g_strdup ("x-canonical-private-button-tint");
	(*out_caps)[11] = g_strdup ("x-canonical-switch-to-application");
	(*out_caps)[12] = g_strdup ("x-canonical-secondary-icon");
  
	/* a temp. compatibility-check for the transition time to allow apps a
	** grace-period to catch up with the capability- and hint-name-changes
	** introduced with notify-osd rev. 224 */
	(*out_caps)[13] = g_strdup ("private-synchronous");
	(*out_caps)[14] = g_strdup ("append");
	(*out_caps)[15] = g_strdup ("private-icon-only");
	(*out_caps)[16] = g_strdup ("truncation");

	(*out_caps)[17] = NULL;

	return TRUE;
}

gboolean
stack_get_server_information (Stack*  self,
			      gchar** out_name,
			      gchar** out_vendor,
			      gchar** out_version,
			      gchar** out_spec_ver)
{
	*out_name     = g_strdup ("notify-osd");
	*out_vendor   = g_strdup ("Canonical Ltd");
	*out_version  = g_strdup ("1.1");
	*out_spec_ver = g_strdup ("1.1");

	return TRUE;
}

gboolean
stack_is_slot_vacant (Stack* self,
                      Slot   slot)
{
	// sanity checks
	if (!self || !IS_STACK (self))
		return FALSE;

	if (slot != SLOT_TOP && slot != SLOT_BOTTOM)
		return FALSE;

	return self->slots[slot] == NULL ? VACANT : OCCUPIED;
}

// return values of -1 for x and y indicate an error by the caller
void
stack_get_slot_position (Stack* self,
                         Slot   slot,
                         gint   bubble_height,
                         gint*  x,
                         gint*  y)
{
	// sanity checks
	if (!x && !y)
		return;

	if (!self || !IS_STACK (self))
	{
		*x = -1;
		*y = -1;
		return;
	}

	if (slot != SLOT_TOP && slot != SLOT_BOTTOM)
	{
		*x = -1;
		*y = -1;
		return;
	}

	// initialize x and y
	defaults_get_top_corner (self->defaults, x, y);

	// differentiate returned top-left corner for top and bottom slot
	// depending on the placement 
	switch (defaults_get_gravity (self->defaults))
	{
		Defaults* d;

		case GRAVITY_EAST:
			d = self->defaults;

			// the position for the sync./feedback bubble
			if (slot == SLOT_TOP)
				*y += defaults_get_desktop_height (d) / 2 -
				      EM2PIXELS (defaults_get_bubble_vert_gap (d) / 2.0f, d) -
				      bubble_height +
				      EM2PIXELS (defaults_get_bubble_shadow_size (d), d);
			// the position for the async. bubble
			else if (slot == SLOT_BOTTOM)
				*y += defaults_get_desktop_height (d) / 2 +
				      EM2PIXELS (defaults_get_bubble_vert_gap (d) / 2.0f, d) -
				      EM2PIXELS (defaults_get_bubble_shadow_size (d), d);
		break;

		case GRAVITY_NORTH_EAST:
			d = self->defaults;

			// there's nothing to do for slot == SLOT_TOP as we
			// already have correct x and y from the call to
			// defaults_get_top_corner() earlier

			// this needs to look at the height of the bubble in the
			// top slot
			if (slot == SLOT_BOTTOM)
			{
				switch (defaults_get_slot_allocation (d))
				{
					case SLOT_ALLOCATION_FIXED:
						*y += EM2PIXELS (defaults_get_icon_size (d), d) +
						      2 * EM2PIXELS (defaults_get_margin_size (d), d) +
						      EM2PIXELS (defaults_get_bubble_vert_gap (d), d); /* +
						      2 * EM2PIXELS (defaults_get_bubble_shadow_size (d), d);*/
					break;

					case SLOT_ALLOCATION_DYNAMIC:
						g_assert (stack_is_slot_vacant (self, SLOT_TOP) == OCCUPIED);
						*y += bubble_get_height (self->slots[SLOT_TOP]) +
						      EM2PIXELS (defaults_get_bubble_vert_gap (d), d) -
						      2 * EM2PIXELS (defaults_get_bubble_shadow_size (d), d);
					break;

					default:
					break;
				}

			}
		break;

		default:
			g_warning ("Unhandled placement!\n");
		break;
	}
}

// call this _before_ the fade-in animation of the bubble starts
gboolean
stack_allocate_slot (Stack*  self,
		     Bubble* bubble,
                     Slot    slot)
{
	// sanity checks
	if (!self || !IS_STACK (self))
		return FALSE;

	if (!bubble || !IS_BUBBLE (bubble))
		return FALSE;

	if (slot != SLOT_TOP && slot != SLOT_BOTTOM)
		return FALSE;

	if (stack_is_slot_vacant (self, slot))
		self->slots[slot] = BUBBLE (g_object_ref ((gpointer) bubble));
	else
		return FALSE;

	return TRUE;
}

// call this _after_ the fade-out animation of the bubble is finished
gboolean
stack_free_slot (Stack*  self,
		 Bubble* bubble)
{
	// sanity checks
	if (!self || !IS_STACK (self))
		return FALSE;

	if (!bubble || !IS_BUBBLE (bubble))
		return FALSE;

	// check top and bottom slots for bubble pointer equality
	if (bubble == self->slots[SLOT_TOP])
	{
		g_object_unref (self->slots[SLOT_TOP]);
		self->slots[SLOT_TOP] = NULL;
	}
	else if (bubble == self->slots[SLOT_BOTTOM])
	{
		g_object_unref (self->slots[SLOT_BOTTOM]);
		self->slots[SLOT_BOTTOM] = NULL;
	}
	else
		return FALSE;

	return TRUE;	
}

void
stack_set_window_thread (Stack*             self,
                         nux::WindowThread* wt)
{
	// sanity checks
	if (!self || !IS_STACK (self))
		return;

	self->wt = wt;
}

const nux::WindowThread*
stack_get_window_thread (Stack* self)
{
	// sanity checks
	if (!self || !IS_STACK (self))
		return NULL;

	return self->wt;
}

void
stack_mark_queue_empty (Stack* self)
{
	// sanity checks
	if (!self || !IS_STACK (self))
		return;

	self->initial_bubble = true;
}

