////////////////////////////////////////////////////////////////////////////////
//3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
//      10        20        30        40        50        60        70        80
//
// notify-osd
//
// test-notification.c - implements unit-tests for exercising API of abstract
//                       notification object
//
// Copyright 2009 Canonical Ltd.
//
// Authors:
//    Mirco "MacSlow" Mueller <mirco.mueller@canonical.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////

#define _XOPEN_SOURCE 500 // needed for usleep() from unistd.h
#include <unistd.h>

#include "notification.h"

static void
test_notification_new (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;
	GTimeVal*     timestamp;

	// create new object
	n = notification_new ();

	// test validity of main notification object
	g_assert (n);

	// test validity of initialized notification object
	g_assert_cmpint (notification_get_id (n), ==, -1);
	g_assert (!notification_get_summary (n));
	g_assert (!notification_get_body (n));
	g_assert_cmpint (notification_get_value (n), ==, -2);
	g_assert (!notification_get_icon_pixbuf (n));
	g_assert_cmpint (notification_get_onscreen_time (n), ==, 0);
	g_assert (!notification_get_sender_name (n));
	g_assert_cmpint (notification_get_sender_pid (n), ==, 0);
	timestamp = notification_get_timestamp (n);
	g_assert_cmpint (timestamp->tv_sec, ==, 0);
	g_assert_cmpint (timestamp->tv_usec, ==, 0);
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_NONE);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_destroy (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// test validity of main notification object
	g_assert (n != NULL);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_id (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if nothing has been set yet it should return -1
	g_assert_cmpint (notification_get_id (n), ==, -1);

	// a negative id should not be stored
	notification_set_id (n, -3);
	g_assert_cmpint (notification_get_id (n), >=, -1);

	// smallest possible id is 0
	notification_set_id (n, 0);
	g_assert_cmpint (notification_get_id (n), ==, 0);

	// largest possible id is G_MAXINT
	notification_set_id (n, G_MAXINT);
	g_assert_cmpint (notification_get_id (n), ==, G_MAXINT);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_summary (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no title has been set yet it should return NULL
	g_assert (notification_get_summary (n) == NULL);

	// set an initial title-text and verify it
	notification_set_summary (n, "Some title text");
	g_assert_cmpstr (notification_get_summary (n), ==, "Some title text");

	// set a new title-text and verify it
	notification_set_summary (n, "The new summary");
	g_assert_cmpstr (notification_get_summary (n), !=, "Some title text");
	g_assert_cmpstr (notification_get_summary (n), ==, "The new summary");

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_body (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no body has been set yet it should return NULL
	g_assert (notification_get_body (n) == NULL);

	// set an initial body-text and verify it
	notification_set_body (n, "Example body text");
	g_assert_cmpstr (notification_get_body (n), ==, "Example body text");

	// set a new body-text and verify it
	notification_set_body (n, "Some new body text");
	g_assert_cmpstr (notification_get_body (n), !=, "Example body text");
	g_assert_cmpstr (notification_get_body (n), ==, "Some new body text");

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_value (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no value has been set yet it should return 0
	g_assert_cmpint (notification_get_value (n), ==, -2);

	// set an initial value and verify it
	notification_set_value (n, 25);
	g_assert_cmpint (notification_get_value (n), ==, 25);

	// set a new value and verify it
	notification_set_value (n, 45);
	g_assert_cmpint (notification_get_value (n), !=, 25);
	g_assert_cmpint (notification_get_value (n), ==, 45);

	// test allowed range
	notification_set_value (n, NOTIFICATION_VALUE_MAX_ALLOWED + 1);
	g_assert_cmpint (notification_get_value (n),
			 ==,
			 NOTIFICATION_VALUE_MAX_ALLOWED);
	notification_set_value (n, NOTIFICATION_VALUE_MIN_ALLOWED - 2);
	g_assert_cmpint (notification_get_value (n),
			 ==,
			 NOTIFICATION_VALUE_MIN_ALLOWED);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_icon_pixbuf (gpointer fixture, gconstpointer user_data)
{
	Notification* n      = NULL;
	GdkPixbuf*    pixbuf = NULL;

	// create new object
	n = notification_new ();

	// if no icon-pixbuf has been set yet it should return NULL
	g_assert (notification_get_icon_pixbuf (n) == NULL);

	// create pixbuf for testing
	pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 100, 100);

	// set an initial icon-pixbuf and verify it
	g_assert (pixbuf);
	notification_set_icon_from_pixbuf (n, pixbuf);
	g_assert (notification_get_icon_pixbuf (n) != NULL);

	// passing an invalid/NULL pointer should not change the stored
	// icon-pixbuf
	notification_set_icon_from_pixbuf (n, NULL);
	g_assert (notification_get_icon_pixbuf (n) != NULL);

	notification_set_icon_from_path (n, "/usr/share/notify-osd/icons/gnome/scalable/status/notification-message-im.svg");
	g_assert (notification_get_icon_pixbuf (n) != NULL);

	notification_set_icon_from_path (n, "/this/will/not/work/icon.png");
	g_assert (notification_get_icon_pixbuf (n) == NULL);

	notification_set_icon (n, "totem");
	g_assert (notification_get_icon_pixbuf (n) != NULL);

	// clean up
	g_object_unref (n);
	n = NULL;

	// more clean up
	g_object_unref (pixbuf);
	pixbuf = NULL;
}

static void
test_notification_setget_onscreen_time (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no onscreen-time has been set yet it should return 0
	g_assert_cmpint (notification_get_onscreen_time (n), ==, 0);

	// setting a negative onscreen-time should fail
	notification_set_onscreen_time (n, -1);
	g_assert_cmpint (notification_get_onscreen_time (n), ==, 0);

	// set an positive onscreen-time and verify it
	notification_set_onscreen_time (n, 1000);
	g_assert_cmpint (notification_get_onscreen_time (n), ==, 1000);

	// set a new onscreen-time and verify it
	notification_set_onscreen_time (n, 5000);
	g_assert_cmpint (notification_get_onscreen_time (n), !=, 1000);
	g_assert_cmpint (notification_get_onscreen_time (n), ==, 5000);

	// setting a new onscreen-time smaller than the currently stored one
	// should fail
	notification_set_onscreen_time (n, 4000);
	g_assert_cmpint (notification_get_onscreen_time (n), ==, 5000);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_dbus_sender (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no dbus-sender has been set yet it should return NULL
	g_assert (notification_get_dbus_sender (n) == NULL);

	// set an initial dbus-sender
	notification_set_dbus_sender (n, "1.345");
	g_assert_cmpstr (notification_get_dbus_sender (n), ==, "1.345");

	// set a new dbus-sender and verify it
	notification_set_dbus_sender (n, "1.4711");
	g_assert_cmpstr (notification_get_dbus_sender (n), !=, "1.345");
	g_assert_cmpstr (notification_get_dbus_sender (n), ==, "1.4711");

	// passing an invalid/NULL pointer should not change the stored
	// sender-name 
	notification_set_dbus_sender (n, NULL);
	g_assert_cmpstr (notification_get_dbus_sender (n), ==, "1.4711");

	// pass an empty (but not NULL) strings
	notification_set_dbus_sender (n, "");
	g_assert_cmpstr (notification_get_dbus_sender (n), ==, "");
	notification_set_dbus_sender (n, "\0");
	g_assert_cmpstr (notification_get_dbus_sender (n), ==, "\0");

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_sender_name (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no sender-name has been set yet it should return NULL
	g_assert (notification_get_sender_name (n) == NULL);

	// set an initial sender-name and verify it
	notification_set_sender_name (n, "evolution");
	g_assert_cmpstr (notification_get_sender_name (n), ==, "evolution");

	// set a new sender-name and verify it
	notification_set_sender_name (n, "pidgin");
	g_assert_cmpstr (notification_get_sender_name (n), !=, "evolution");
	g_assert_cmpstr (notification_get_sender_name (n), ==, "pidgin");

	// passing an invalid/NULL pointer should not change the stored
	// sender-name 
	notification_set_sender_name (n, NULL);
	g_assert_cmpstr (notification_get_sender_name (n), ==, "pidgin");

	// pass an empty (but not NULL) strings
	notification_set_sender_name (n, "");
	g_assert_cmpstr (notification_get_sender_name (n), ==, "");
	notification_set_sender_name (n, "\0");
	g_assert_cmpstr (notification_get_sender_name (n), ==, "\0");

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_sender_pid (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if nothing has been set yet it should return 0
	g_assert_cmpint (notification_get_sender_pid (n), ==, 0);

	// a negative pid makes no sense and should therefore not be stored
	notification_set_sender_pid (n, -1);
	g_assert_cmpint (notification_get_sender_pid (n), ==, 0);

	// smallest possible pid is 1
	notification_set_sender_pid (n, 1);
	g_assert_cmpint (notification_get_sender_pid (n), ==, 1);

	// largest possible id is G_MAXINT
	notification_set_sender_pid (n, G_MAXINT);
	g_assert_cmpint (notification_get_sender_pid (n), ==, G_MAXINT);

	// a pid of 0 would mean something before the init process is sending us
	// a notification, leave the stored pid untouched
	notification_set_sender_pid (n, 0);
	g_assert_cmpint (notification_get_sender_pid (n), ==, G_MAXINT);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_timestamp (gpointer fixture, gconstpointer user_data)
{
	Notification* n     = NULL;
	GTimeVal*       tvptr = NULL;
	GTimeVal        tv_old;
	GTimeVal        tv_new;

	// create new object
	n = notification_new ();

	// if no reception-time has been set yet it should return 0/0
	tvptr = notification_get_timestamp (n);
	g_assert_cmpint (tvptr->tv_sec, ==, 0);
	g_assert_cmpint (tvptr->tv_usec, ==, 0);

	// ehm... well, get current time
	g_get_current_time (&tv_old);

	// store current time as reception-time and verify it
	notification_set_timestamp (n, &tv_old);
	tvptr = notification_get_timestamp (n);
	g_assert_cmpint (tvptr->tv_sec, ==, tv_old.tv_sec);
	g_assert_cmpint (tvptr->tv_usec, ==, tv_old.tv_usec);

	// wait at least two seconds
	sleep (2);

	// get current time
	g_get_current_time (&tv_new);

	// trying to store an older timestamp over a newer one should fail
	// second-granularity
	notification_set_timestamp (n, &tv_new);
	notification_set_timestamp (n, &tv_old);
	tvptr = notification_get_timestamp (n);
	g_assert_cmpint (tvptr->tv_sec, !=, tv_old.tv_sec);
	g_assert_cmpint (tvptr->tv_sec, ==, tv_new.tv_sec);

	// get current time
	g_get_current_time (&tv_old);
	notification_set_timestamp (n, &tv_old);

	// wait some micro-seconds
	usleep (10000);

	// get current time
	g_get_current_time (&tv_new);

	// trying to store an older timestamp over a newer one should fail
	// microsecond-granularity
	notification_set_timestamp (n, &tv_new);
	notification_set_timestamp (n, &tv_old);
	tvptr = notification_get_timestamp (n);
	g_assert_cmpint (tvptr->tv_usec, !=, tv_old.tv_usec);
	g_assert_cmpint (tvptr->tv_usec, ==, tv_new.tv_usec);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_urgency (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no urgency has been set yet it should return urgency-none
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_NONE);

	// test all three urgency-levels
	notification_set_urgency (n, URGENCY_HIGH);
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_HIGH);
	notification_set_urgency (n, URGENCY_LOW);
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_LOW);
	notification_set_urgency (n, URGENCY_NORMAL);
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_NORMAL);

	// test non-urgency levels, last valid urgency should be returned
	notification_set_urgency (n, (Urgency) 5);
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_NORMAL);
	notification_set_urgency (n, (Urgency) 23);
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_NORMAL);
	notification_set_urgency (n, (Urgency) -2);
	g_assert_cmpint (notification_get_urgency (n), ==, URGENCY_NORMAL);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_actions (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;
	gchar* valid_actions1[] = {"Accept",
		                       "Decline",
                               "Away",
                               "Busy",
                               NULL};
	gchar* valid_actions2[] = {"Ok",
                               "Cancel",
                               "Screw you",
                               "Bugger off",
                               "Feel ignored",
                               NULL};

	// create new object
	n = notification_new ();

	// fresh notification should have no actions at all
	g_assert (notification_get_actions (n) == NULL);

	// set some sample actions and check them being set correctly
	notification_set_actions (n, valid_actions1);
	g_assert (notification_get_actions (n) != NULL);
	g_assert_cmpint (notification_get_num_actions (n), ==, 4);
	g_assert_cmpstr (notification_get_action_by_index (n, 0), ==, "Accept");
	g_assert_cmpstr (notification_get_action_by_index (n, 1), ==, "Decline");
	g_assert_cmpstr (notification_get_action_by_index (n, 2), ==, "Away");
	g_assert_cmpstr (notification_get_action_by_index (n, 3), ==, "Busy");

	// set another set of sample actions and check them being set correctly
	notification_set_actions (n, valid_actions2);
	g_assert (notification_get_actions (n) != NULL);
	g_assert_cmpint (notification_get_num_actions (n), ==, 5);
	g_assert_cmpstr (notification_get_action_by_index (n, 0), ==, "Ok");
	g_assert_cmpstr (notification_get_action_by_index (n, 1), ==, "Cancel");
	g_assert_cmpstr (notification_get_action_by_index (n, 2), ==, "Screw you");
	g_assert_cmpstr (notification_get_action_by_index (n, 3), ==, "Bugger off");
	g_assert_cmpstr (notification_get_action_by_index (n, 4), ==, "Feel ignored");

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_visible (gpointer fixture, gconstpointer user_data)
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	g_assert_cmpint (notification_is_visible (n), ==, FALSE);
	notification_set_visible (n);
	g_assert_cmpint (notification_is_visible (n), ==, TRUE);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_button_tint ()
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	g_assert_cmpint (notification_get_button_tint (n), ==, FALSE);
	notification_set_button_tint (n, TRUE);
	g_assert_cmpint (notification_get_button_tint (n), ==, TRUE);
	notification_set_button_tint (n, FALSE);
	g_assert_cmpint (notification_get_button_tint (n), ==, FALSE);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_append ()
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	g_assert_cmpint (notification_get_append_hint (n), ==, FALSE);
	notification_set_append_hint (n, TRUE);
	g_assert_cmpint (notification_get_append_hint (n), ==, TRUE);
	notification_set_append_hint (n, FALSE);
	g_assert_cmpint (notification_get_append_hint (n), ==, FALSE);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_sync ()
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	g_assert_cmpint (notification_get_sync_hint (n), ==, FALSE);
	notification_set_sync_hint (n, TRUE);
	g_assert_cmpint (notification_get_sync_hint (n), ==, TRUE);
	notification_set_sync_hint (n, FALSE);
	g_assert_cmpint (notification_get_sync_hint (n), ==, FALSE);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_icon_only ()
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	g_assert_cmpint (notification_get_icon_only_hint (n), ==, FALSE);
	notification_set_icon_only_hint (n, TRUE);
	g_assert_cmpint (notification_get_icon_only_hint (n), ==, TRUE);
	notification_set_icon_only_hint (n, FALSE);
	g_assert_cmpint (notification_get_icon_only_hint (n), ==, FALSE);

	// clean up
	g_object_unref (n);
	n = NULL;
}

static void
test_notification_setget_secondary_icon ()
{
	Notification* n = NULL;

	// create new object
	n = notification_new ();

	// if no secondary icon has been set yet, it should return NULL
	g_assert (notification_get_secondary_icon (n) == NULL);

	// set an initial secondary icon and verify it
	notification_set_secondary_icon_from_path (n, "/usr/share/notify-osd/icons/hicolor/scalable/status/notification-message-im.svg");
	g_assert (notification_get_secondary_icon (n) != NULL);

	// passing an invalid/NULL pointer should not change the stored icon-pixbuf
	notification_set_secondary_icon_from_path (n, NULL);
	g_assert (notification_get_icon_pixbuf (n) != NULL);

	// set different secondary icon
	notification_set_secondary_icon_from_path (n, "/usr/share/notify-osd/icons/hicolor/scalable/status/notification-message-email.svg");
	g_assert (notification_get_secondary_icon (n) != NULL);

	// failure test
	notification_set_secondary_icon_from_path (n, "/this/will/not/work/icon.png");
	g_assert (notification_get_secondary_icon (n) == NULL);

	// clean up
	g_object_unref (n);
	n = NULL;
}

GTestSuite *
test_notification_create_test_suite (void)
{
	GTestSuite *ts = NULL;
	GTestCase  *tc = NULL;

	ts = g_test_create_suite ("notification");
	tc = g_test_create_case ("can create",
				 0,
				 NULL,
				 NULL,
				 (GTestFixtureFunc) test_notification_new,
				 NULL);
	g_test_suite_add (ts, tc);

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can destroy",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_destroy,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get id",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_id,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get summary",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_summary,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get body",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_body,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get value",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_value,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get icon-pixbuf",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_icon_pixbuf,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get onscreen-time",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_onscreen_time,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get dbus-sender",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_dbus_sender,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get sender-name",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_sender_name,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get sender-pid",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_sender_pid,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get timestamp",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_timestamp,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get urgency",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_urgency,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get actions",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_actions,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get visible",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_visible,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get button-tint",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_button_tint,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get append",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_append,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get sync",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_sync,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get icon-only",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_icon_only,
			NULL));

	g_test_suite_add (
		ts,
		g_test_create_case (
			"can set|get secondary-icon",
			0,
			NULL,
			NULL,
			(GTestFixtureFunc) test_notification_setget_secondary_icon,
			NULL));

	return ts;
}
